"""
Create a NeXus file for time-resolved collections on I19-2 using parameters passed from GDA.
"""
from __future__ import annotations

import logging
from collections import namedtuple
from pathlib import Path
from typing import Tuple

import numpy as np
from numpy.typing import DTypeLike

from .. import log
from ..nxs_utils import (
    Attenuator,
    Beam,
    Detector,
    EigerDetector,
    Goniometer,
    Source,
    TristanDetector,
)
from ..nxs_utils.Detector import DetectorType, UnknownDetectorTypeError
from ..nxs_utils.ScanUtils import calculate_scan_points
from ..nxs_write.NXmxWriter import EventNXmxFileWriter, NXmxFileWriter
from ..utils import get_iso_timestamp, get_nexus_filename
from .beamline_utils import BeamlineAxes, collection_summary_log
from .GDAtools.ExtendedRequest import (
    ExtendedRequestIO,
    read_det_position_from_xml,
    read_scan_from_xml,
)
from .GDAtools.GDAjson2params import JSONParamsIO

# Define a logger object and a formatter
logger = logging.getLogger("nexgen.I19-2_NeXus_gda")

tr_collect = namedtuple(
    "tr_collect",
    [
        "meta_file",
        "xml_file",
        "detector_name",
        "exposure_time",
        "wavelength",
        "beam_center",
        "start_time",
        "stop_time",
        "geometry_json",  # Define these 2 as None
        "detector_json",
    ],
)

tr_collect.__doc__ = (
    """Information extracted from GDA containing collection parameters."""
)
tr_collect.meta_file.__doc__ = "Path to _meta.h5 file."
tr_collect.xml_file.__doc__ = "Path to GDA-generated xml file."
tr_collect.detector_name.__doc__ = "Name of the detector in use for current experiment."
tr_collect.exposure_time.__doc__ = "Exposure time, in s."
tr_collect.wavelength.__doc__ = "Incident beam wavelength, in A."
tr_collect.beam_center.__doc__ = "Beam center (x,y) position, in pixels."
tr_collect.start_time.__doc__ = "Collection start time."
tr_collect.stop_time.__doc__ = "Collection end time."
tr_collect.geometry_json.__doc__ = (
    "Path to GDA-generated JSON file describing the beamline geometry."
)
tr_collect.detector_json.__doc__ = (
    "Path to GDA-generated JSON file describing the detector."
)


def tristan_writer(
    master_file: Path,
    TR: namedtuple,
    axes_params: BeamlineAxes,
    det_params: DetectorType,
    timestamps: Tuple[str, str] = (None, None),
):
    """
    A function to call the nexus writer for Tristan 10M detector.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (namedtuple): Parameters passed from the beamline.
        axes_params (BeamlineAxes): Axes for goniometer, detector and detector module.
        det_params (DetectorType): Detector definition for Tristan.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to None.
    """
    ecr = ExtendedRequestIO(TR.xml_file)
    # Read information from xml file
    logger.info("Read xml file.")
    scan_axis, pos, _ = read_scan_from_xml(ecr)
    # n_Frames is only useful for eiger
    # pos[scan_axis][::-1] is scan range
    scan_range = pos[scan_axis][:-1]
    # Define OSC scans dictionary
    OSC = {scan_axis: scan_range}

    det_positions = read_det_position_from_xml(ecr, det_params.description)

    # Attenuator
    attenuator = Attenuator(ecr.getTransmission())

    # Beam
    beam = Beam(TR.wavelength)

    # Source
    source = Source("I19-2")

    # Detector
    det_axes = axes_params.det_axes
    det_axes[0].start_pos = det_positions[0]  # two_theta
    det_axes[1].start_pos = det_positions[1]  # det_z
    detector = Detector(
        det_params,
        det_axes,
        TR.beam_center,
        TR.exposure_time,
        [axes_params.fast_axis, axes_params.slow_axis],
    )

    # Goniometer
    gonio_axes = axes_params.gonio
    for k, v in pos.items():
        # Get correct start positions
        idx = [n for n, ax in enumerate(gonio_axes) if ax.name == k][0]
        gonio_axes[idx].start_pos = v[0]
    goniometer = Goniometer(gonio_axes, OSC)

    collection_summary_log(
        logger,
        goniometer,
        detector,
        attenuator,
        beam,
        source,
        timestamps,
    )

    # Get on with the writing now...
    try:
        EventFileWriter = EventNXmxFileWriter(
            master_file,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
        )
        EventFileWriter.write(start_time=timestamps[0])
        if timestamps[1]:
            EventFileWriter.update_timestamps(timestamps[1], "end_time")
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )


def eiger_writer(
    master_file: Path,
    TR: namedtuple,
    axes_params: BeamlineAxes,
    det_params: DetectorType,
    timestamps: Tuple[str, str] = (None, None),
    vds_dtype: DTypeLike = np.uint16,
):
    """
    A function to call the nexus writer for Eiger 2X 4M detector.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (namedtuple): Parameters passed from the beamline.
        axes_params (BeamlineAxes): Axes for goniometer, detector and detector module.
        det_params (DetectorType): Detector definition for Eiger.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to (None, None).
        vds_dtype (DtypeLike): Data type for vds as np.uint##.
    """
    ecr = ExtendedRequestIO(TR.xml_file)
    # Read information from xml file
    logger.info("Read xml file.")
    scan_axis, pos, n_frames = read_scan_from_xml(ecr)

    det_positions = read_det_position_from_xml(ecr, det_params.description)

    # Attenuator
    attenuator = Attenuator(ecr.getTransmission())

    # Beam
    beam = Beam(TR.wavelength)

    # Source
    source = Source("I19-2")

    # Detector
    det_axes = axes_params.det_axes
    det_axes[0].start_pos = det_positions[0]  # two_theta
    det_axes[1].start_pos = det_positions[1]  # det_z
    detector = Detector(
        det_params,
        det_axes,
        TR.beam_center,
        TR.exposure_time,
        [axes_params.fast_axis, axes_params.slow_axis],
    )

    # Goniometer
    gonio_axes = axes_params.gonio
    for k, v in pos.items():
        # Get correct start positions
        idx = [n for n, ax in enumerate(gonio_axes) if ax.name == k][0]
        gonio_axes[idx].start_pos = v[0]

    # Get scan range array
    logger.info("Calculating scan range...")
    scan_idx = [n for n, ax in enumerate(gonio_axes) if ax.name == scan_axis][0]

    # Define OSC scans dictionary
    OSC = calculate_scan_points(
        gonio_axes[scan_idx], rotation=True, tot_num_imgs=n_frames
    )

    goniometer = Goniometer(gonio_axes, OSC)

    collection_summary_log(
        logger,
        gonio_axes,
        [scan_axis],
        detector,
        attenuator,
        beam,
        source,
        timestamps,
    )

    # Get on with the writing now...
    try:
        NXmx_Writer = NXmxFileWriter(
            master_file,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
            n_frames,
        )
        NXmx_Writer.write(start_time=timestamps[0])
        if timestamps[1]:
            NXmx_Writer.update_timestamps(timestamps[1], "end_time")
        NXmx_Writer.write_vds(
            vds_shape=(n_frames, *detector.detector_params.image_size),
            vds_dtype=vds_dtype,
        )
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )


def write_nxs(**tr_params):
    """
    Gather all parameters from the beamline and call the NeXus writers.

    Keyword Args:
        meta_file (Path | str): Path to _meta.h5 file.
        xml_file (Path | str): Path to gda-generated xml file.
        detector_name (str): Detector in use.
        exposure_time (float): Exposure time, in s.
        wavelength (float): Wavelength of incident beam, in A.
        beam_center (List[float, float]): Beam center position, in pixels.
        start_time (datetime): Experiment start time.
        stop_time (datetime): Experiment end time.
        geometry_json (Path | str): Path to GDA generated geometry json file.
        detector_json (Path | str): Path to GDA generated detector json file.
        vds_dtype (DtypeLike): Data type for vds as np.uint##.
    """
    # Get info from the beamline
    TR = tr_collect(
        meta_file=Path(tr_params["meta_file"]).expanduser().resolve(),
        xml_file=Path(tr_params["xml_file"]).expanduser().resolve(),
        detector_name=tr_params["detector_name"],
        exposure_time=tr_params["exposure_time"],
        wavelength=tr_params["wavelength"],
        beam_center=tr_params["beam_center"],
        start_time=tr_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")  #
        if tr_params["start_time"]
        else None,  # This should be datetiem type
        stop_time=tr_params["stop_time"].strftime(
            "%Y-%m-%dT%H:%M:%S"
        )  # .strftime("%Y-%m-%dT%H:%M:%S")
        if tr_params["stop_time"]
        else None,  # idem.
        geometry_json=tr_params["geometry_json"]
        if tr_params["geometry_json"]
        else None,
        detector_json=tr_params["detector_json"]
        if tr_params["detector_json"]
        else None,
    )

    # Define a file handler
    logfile = TR.meta_file.parent / "nexus_writer.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info("NeXus file writer for beamline I19-2 at DLS.")
    logger.info(f"Detector in use for this experiment: {TR.detector_name}.")
    logger.info(f"Current collection directory: {TR.meta_file.parent}")
    # Add some information to logger
    logger.info("Creating a NeXus file for %s ..." % TR.meta_file.name)
    # Get NeXus filename
    master_file = get_nexus_filename(TR.meta_file)
    logger.info("NeXus file will be saved as %s" % master_file)

    # Get some parameters in here
    if "eiger" in TR.detector_name.lower():
        from .I19_2_params import I19_2Eiger as axes_params

        det_params = EigerDetector(
            "Eiger 2X 4M",
            (2162, 2068),
            "CdTe",
            50649,
            -1,
        )
    elif "tristan" in TR.detector_name.lower():
        from .I19_2_params import I19_2Tristan as axes_params

        det_params = TristanDetector("Tristan 10M", (3043, 4183))
    else:
        raise UnknownDetectorTypeError("Unknown detector name passed.")

    # Get goniometer and detector parameters
    if TR.geometry_json:
        logger.info("Reading geometry from json file.")
        gonio_axes = JSONParamsIO(TR.geometry_json).get_goniometer_axes_from_file()
        det_axes = JSONParamsIO(TR.geometry_json).get_detector_axes_from_file()
        # Overwrite
        axes_params.gonio = gonio_axes
        axes_params.det_axes = det_axes

    if TR.detector_json:
        logger.info("Reading detector parameters from json file.")
        det_params = JSONParamsIO(TR.detector_json).get_detector_params_from_file()
        fast_axis, slow_axis = JSONParamsIO(
            TR.detector_json
        ).get_fast_and_slow_direction_vectors_from_file(det_params.description)
        # Overwrite
        axes_params.fast_axis = fast_axis
        axes_params.slow_axis = slow_axis

    # Get timestamps in the correct format if they aren't already
    timestamps = (
        get_iso_timestamp(TR.start_time),
        get_iso_timestamp(TR.stop_time),
    )

    if "tristan" in TR.detector_name:
        tristan_writer(master_file, TR, axes_params, det_params, timestamps)
    else:
        vds_dtype = (
            np.uint16 if "vds_dtype" not in tr_params.keys() else tr_params["vds_dtype"]
        )
        eiger_writer(master_file, TR, axes_params, det_params, timestamps, vds_dtype)



"""
Command line tool to get an existing .phil file with goniometer/detector metadata or to create a new one.
These files can be used as input for the NeXus generator CLI.
"""

import argparse
import logging
import shutil
import sys
from pathlib import Path

import freephil

from .. import log, templates
from . import config_parser, nexus_parser, version_parser

try:
    from importlib.resources import files
except ImportError:
    # Python < 3.9 compatibility
    from importlib_resources import files


# Define a logger object
logger = logging.getLogger("nexgen.NeXusGenerator")

scopes = freephil.parse(
    """
    include scope nexgen.command_line.nxs_phil.goniometer_scope
    include scope nexgen.command_line.nxs_phil.instrument_scope
    include scope nexgen.command_line.nxs_phil.detector_scope
    include scope nexgen.command_line.nxs_phil.module_scope
    """,
    process_includes=True,
)

parser = argparse.ArgumentParser(description=__doc__, parents=[version_parser])
parser.add_argument("--debug", action="store_const", const=True)


def list_available_phil():
    filedir = files(templates)
    for f in filedir.glob("*.phil"):
        print(f.name)


def get_beamline_phil(args):
    # Determine where to save file
    if args.output:
        odir = Path(args.output).expanduser().resolve()
    else:
        odir = Path(".").expanduser().resolve()
        logger.warning(
            "No output directory was specified by user. A copy of the file will be saved in the current directory."
        )

    # Look for file
    filedir = files(templates)
    found = [f for f in sorted(filedir.glob("*.phil")) if args.phil_file == f.name]
    if len(found) == 0:
        logger.info(f"No {args.phil_file} found.")
    elif len(found) == 1:
        logger.info(f"{args.phil_file} found. Copying in {odir}.")
        shutil.copy(found[0], odir)


def create_new_phil(args):
    cl = scopes.command_line_argument_interpreter()
    working_phil = scopes.fetch(cl.process_and_fetch(args.phil_args))

    if args.show_config:
        working_phil.show(attributes_level=args.attributes_level)
        sys.exit()

    if args.skip is True:
        logger.warning(
            "The scope containing Tristan specs will be removed before writing."
        )
        tristanSpec = [
            obj for obj in working_phil.objects if "tristanSpec" in obj.as_str()
        ][0]
        working_phil.objects.remove(tristanSpec)

    # Write to file
    if args.filename:
        filename = Path(args.filename).expanduser().resolve()
        logger.info(f"Writing new .phil file to {filename.name} in {filename.parent}.")
        with open(filename, "w") as fout:
            fout.write(working_phil.as_str())
    else:
        logger.info(working_phil.as_str())


# Define subparsers
subparser = parser.add_subparsers(
    help="Run nexgen_phil <command> --help to see the options for each command.",
    required=True,
    dest="command",
)

parser_list = subparser.add_parser(
    "list",
    description=("Print out a list of currently available .phil template files."),
)
parser_list.set_defaults(func=list_available_phil)

parser_get = subparser.add_parser(
    "get",
    description=(
        "Get a copy of a .phil file to use as input for the NeXus file writer."
    ),
)
parser_get.add_argument("phil_file", type=str, help="Requested file name.")
parser_get.add_argument("-o", "--output", type=str, help="Specify output directory.")
parser_get.set_defaults(func=get_beamline_phil)

parser_create = subparser.add_parser(
    "new",
    description=("Write a new .phil file."),
    parents=[nexus_parser, config_parser],
)
parser_create.add_argument(
    "-f", "--filename", type=str, help="Filename for new .phil template."
)
parser_create.add_argument(
    "-s",
    "--skip",
    help="Pass to avoid writing tristanSpec scope for non-Tristan detectors.",
    action="store_true",
)
parser_create.set_defaults(func=create_new_phil)


def main():
    # Configure logging
    log.config()

    args = parser.parse_args()
    if args.command == "list":
        args.func()
    else:
        args.func(args)


# main()



"""
Tools to get the information stored inside the _meta.h5 file and overwrite the phil scope.
"""

from __future__ import annotations

import logging
from typing import Dict, List

import h5py
import numpy as np
from numpy.typing import DTypeLike

from ..nxs_utils import Axis
from ..utils import ScopeExtract, units_of_length
from .Metafile import DectrisMetafile, TristanMetafile

# TODO actually define the type for scope extract and replace Any with Union
overwrite_logger = logging.getLogger("nexgen.MetaReader")
overwrite_logger.setLevel(logging.DEBUG)


def overwrite_beam(meta_file: h5py.File, name: str, beam: Dict | ScopeExtract):
    """
    Looks for the wavelength value in the _meta.h5 file.
    If found, it overwrites the value that was parsed from the command line.

    Args:
        meta_file (h5py.File): Handle for _meta.h5 file.
        name (str): Detector description.
        beam (Dict | Any): Scope extract or dictionary defining the beam.

    Raises:
        ValueError: If an invalid detector description is passed. Allowed detectors for this function: Eiger.
    """
    if "eiger" in name.lower():
        meta = DectrisMetafile(meta_file)
        wl = meta.get_wavelength()
        if wl is None:
            overwrite_logger.info("No wavelength information found in meta file.")
            return
    else:
        raise ValueError("Unknown detector: please pass a valid detector description.")
    # If value exists, overwrite. Otherwise, create.
    overwrite_logger.warning("Wavelength will be overwritten.")
    overwrite_logger.info(f"Value for wavelength found in meta file: {wl}")
    if isinstance(beam, dict):
        beam["wavelength"] = wl
    else:
        try:
            beam.__dict__["wavelength"] = wl
        except KeyError:
            beam.__inject__("wavelength", wl)


def overwrite_detector(
    meta_file: h5py.File, detector: Dict | ScopeExtract, ignore: List = None
) -> List:
    """
    Looks through the _meta.h5 file for informtion relating to NXdetector.

    Args:
        meta_file (h5py.File): Handle for _meta.h5 file.
        detector (Dict | Any): Scope extract or dictionary defining the detector.
        ignore (List, optional): List of datasets that should not be overwritten by the meta file. Defaults to None.

    Raises:
        ValueError: If an invalid detector description is passed. Allowed detectors for this function: Eiger, Tristan.

    Returns:
        link_list (List): A list of elements to be linked instead of copied in the NeXus file.
    """
    new_values = {}
    link_list = [[], []]
    if isinstance(detector, dict):
        detector_name = detector["description"].lower()
    else:
        detector_name = detector.description.lower()
    if "tristan" in detector_name:
        meta = TristanMetafile(meta_file)
        new_values["n_modules"] = meta.find_number_of_modules()
        new_values["meta_version"] = meta.find_meta_version()
        link_list[1].append("meta_version")
        overwrite_logger.info("Looking through meta file for Tristan detector.")
        overwrite_logger.info("Number of modules: %d" % new_values["n_modules"])
        overwrite_logger.info(
            "Found meta_version located at: %s " % new_values["meta_version"]
        )
    elif "eiger" in detector_name:
        meta = DectrisMetafile(meta_file)
        overwrite_logger.info("Looking through meta file for Eiger detector.")
        if meta.hasMask is True:
            overwrite_logger.info("Mask has been located in meta file")
            mask_info = meta.find_mask()
            new_values["pixel_mask"] = mask_info[0]
            new_values["pixel_mask_applied"] = mask_info[1]
            link_list[0].append("pixel_mask")
            link_list[0].append("pixel_mask_applied")
        if meta.hasFlatfield is True:
            overwrite_logger.info("Flatfield has been located in meta file")
            flatfield_info = meta.find_flatfield()
            new_values["flatfield"] = flatfield_info[0]
            new_values["flatfield_applied"] = flatfield_info[1]
            link_list[0].append("flatfield")
            link_list[0].append("flatfield_applied")
        if meta.hasDectrisGroup is True:
            new_values["software_version"] = meta.find_software_version()
            new_values["threshold_energy"] = meta.find_threshold_energy()
            new_values["bit_depth_readout"] = meta.find_bit_depth_readout()
            new_values["detector_readout_time"] = meta.find_detector_readout_time()
            new_values["serial_number"] = meta.find_detector_number()
            link_list[0].append("threshold_energy")
            link_list[0].append("bit_depth_readout")
            link_list[0].append("detector_readout_time")
            link_list[0].append("serial_number")
            link_list[1].append("software_version")
            pix = meta.get_pixel_size()
            new_values["pixel_size"] = [
                units_of_length(pix[0]),
                units_of_length(pix[1]),
            ]
            overwrite_logger.warning("Pixel_size will be overwritten.")
            overwrite_logger.info(
                f"Values for x and y pixel size found in meta file: {pix[0]}, {pix[1]}"
            )
            new_values["beam_center"] = meta.get_beam_center()
            overwrite_logger.warning("Beam_center will be overwritten.")
            overwrite_logger.info(
                f"Values for x and y beam center position found in meta file: "
                f"{new_values['beam_center']}"
            )
            sensor_info = meta.get_sensor_information()
            new_values["sensor_material"] = sensor_info[0]
            overwrite_logger.warning("Sensor material will be overwritten.")
            overwrite_logger.info(
                f"Value for sensor material found in meta file: {sensor_info[0]}"
            )
            new_values["sensor_thickness"] = units_of_length(sensor_info[1])
            overwrite_logger.warning("Sensor thickness will be overwritten.")
            overwrite_logger.info(
                f"Value for sensor thickness found in meta file: {sensor_info[1]}"
            )
            new_values["overload"] = meta.get_saturation_value()
            overwrite_logger.warning("Saturation value (overload) will be overwritten.")
            overwrite_logger.info(
                f"Value for overload found in meta file: {new_values['overload']}"
            )
    else:
        overwrite_logger.warning("Unknown detector, exit.")
        raise ValueError("Unknown detector: please pass a valid detector description.")

    if ignore:
        overwrite_logger.warning(
            f"The following datasets are not going to be overwritten: {ignore}"
        )
        for i in ignore:
            if i in new_values:
                del new_values[i]

    for k, v in new_values.items():
        if isinstance(detector, dict):
            detector[k] = v
        else:
            try:
                detector.__dict__[k] = v
            except KeyError:
                detector.__inject__(k, v)
    return link_list


def define_vds_data_type(meta_file: DectrisMetafile) -> DTypeLike:
    """Define the data type for the VDS from the bit_depth defined in the meta file.

    Args:
        meta_file (DectrisMetafile): Handle to Dectris-shaped meta.h5 file.

    Returns:
        DTypeLike: Data type as np.uint##.
    """
    overwrite_logger.info("Define dtype for VDS creating from bit_depth_image.")
    # meta = DectrisMetafile(meta_file)

    nbits = meta_file.get_bit_depth_image()
    overwrite_logger.info(f"Found value for bit_depth_image: {nbits}.")
    if nbits == 32:
        return np.uint32
    elif nbits == 8:
        return np.uint8
    else:
        return np.uint16


def update_axes_from_meta(
    meta_file: DectrisMetafile,
    axes_list: List[Axis],
    osc_axis: str | None = None,
    use_config: bool = False,
):
    """Update goniometer or detector axes values from those stores in the _dectris group.

    Args:
        meta_file (DectrisMetafile): Handle to Dectris-shaped meta.h5 file.
        axes_list (List[Axis]): List of axes to look up and eventually update.
        osc_axis (str | None, optional): If passed, the number of images corresponding to the osc_axis \
            will be updated too. Defaults to None.
        use_config (bool, optional): If passed read from config dataset in meta file instead of _dectris\
            group. Defaults to False.
    """
    overwrite_logger.info("Updating axes list with values saved to _dectris group.")
    if meta_file.hasDectrisGroup is False:
        overwrite_logger.warning(
            "No Dectris group in meta file. No values will be updated."
        )
        return

    if use_config is True and meta_file.hasConfig is True:
        config = meta_file.read_config_dset()
    else:
        config = meta_file.read_dectris_config()
    num = meta_file.get_full_number_of_images()

    for ax in axes_list:
        if f"{ax.name}_start" in config.keys():
            ax.start_pos = config[f"{ax.name}_start"]
            overwrite_logger.info(f"Start value for axis {ax.name}: {ax.start_pos}.")
            if f"{ax.name}_increment" in config.keys():
                ax.increment = config[f"{ax.name}_increment"]
                overwrite_logger.info(
                    f"Increment value for axis {ax.name}: {ax.increment}."
                )
        if osc_axis and ax.name == osc_axis:
            ax.num_steps = num

        if ax.name == "det_z":
            dist = units_of_length(meta_file.get_detector_distance())
            ax.start_pos = dist.to("mm").magnitude
            overwrite_logger.info(f"Start value for axis {ax.name}: {ax.start_pos}.")



from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from nexgen.nxs_utils import (
    Axis,
    EigerDetector,
    Facility,
    Goniometer,
    Source,
    TransformationType,
    TristanDetector,
)
from nexgen.nxs_write.NXclassWriters import (
    write_NXcollection,
    write_NXcoordinate_system_set,
    write_NXdata,
    write_NXdatetime,
    write_NXdetector,
    write_NXdetector_module,
    write_NXentry,
    write_NXinstrument,
    write_NXnote,
    write_NXsample,
    write_NXsource,
)

test_module = {"fast_axis": [1, 0, 0], "slow_axis": [0, 1, 0]}


def test_given_no_data_files_when_write_NXdata_then_assert_error():
    mock_hdf5_file = MagicMock()
    with pytest.raises(OSError):
        write_NXdata(mock_hdf5_file, [], [], "", "", [])


def test_write_NXentry(dummy_nexus_file):
    entry = write_NXentry(dummy_nexus_file)

    assert dummy_nexus_file["/entry/"].attrs["NX_class"] == b"NXentry"
    assert dummy_nexus_file["/entry/"].attrs["default"] == b"data"
    assert dummy_nexus_file["/entry/"].attrs["version"] == b"1.0"

    assert "definition" in entry.keys()
    assert dummy_nexus_file["/entry/definition"][()] == b"NXmx"


def test_write_NXSource(dummy_nexus_file, mock_source):
    write_NXsource(dummy_nexus_file, mock_source)

    assert dummy_nexus_file["/entry/source/name"][()] == b"Diamond Light Source"
    assert dummy_nexus_file["/entry/source/name"].attrs["short_name"] == b"DLS"
    assert "probe" not in dummy_nexus_file["/entry/source"].keys()


def test_write_NXSource_with_probe(dummy_nexus_file, mock_source):
    mock_source.probe = "electron"
    write_NXsource(dummy_nexus_file, mock_source)

    assert dummy_nexus_file["/entry/source/probe"][()] == b"electron"


def test_given_no_data_type_specified_when_write_NXdata_then_exception_raised(
    dummy_nexus_file, mock_goniometer
):
    with pytest.raises(ValueError):
        write_NXdata(
            dummy_nexus_file,
            [Path("tmp")],
            mock_goniometer.axes_list,
            "",
            mock_goniometer.scan,
        )


def test_given_one_data_file_when_write_NXdata_then_data_in_file(
    dummy_nexus_file, mock_goniometer
):
    write_NXdata(
        dummy_nexus_file,
        [Path("tmp")],
        mock_goniometer.axes_list,
        "images",
        mock_goniometer.scan,
    )
    assert dummy_nexus_file["/entry/data"].attrs["NX_class"] == b"NXdata"
    assert "data_000001" in dummy_nexus_file["/entry/data"]


def test_given_scan_axis_when_write_NXdata_then_axis_in_data_entry_with_correct_data_and_attributes(
    dummy_nexus_file, mock_goniometer
):
    test_axis = "omega"
    test_scan_range = np.arange(0, 90, 1)
    axis_entry = f"/entry/data/{test_axis}"

    write_NXdata(
        dummy_nexus_file,
        [Path("tmp")],
        mock_goniometer.axes_list,
        "images",
        mock_goniometer.scan,
    )

    assert test_axis in dummy_nexus_file["/entry/data"]
    assert_array_equal(test_scan_range, dummy_nexus_file[axis_entry][:])
    assert dummy_nexus_file[axis_entry].attrs["depends_on"] == b"."
    assert dummy_nexus_file[axis_entry].attrs["transformation_type"] == b"rotation"
    assert dummy_nexus_file[axis_entry].attrs["units"] == b"deg"
    assert_array_equal(dummy_nexus_file[axis_entry].attrs["vector"][:], [-1.0, 0.0, 0])


def test_given_scan_axis_when_write_NXsample_then_scan_axis_data_copied_from_data_group_as_well_as_increment_set_and_end(
    dummy_nexus_file, mock_goniometer
):
    test_axis = "omega"
    test_scan_range = [0, 1, 2]
    axis_entry = f"/entry/sample/sample_{test_axis}/{test_axis}"
    osc_scan = {test_axis: test_scan_range}

    # Doing this to write the scan axis data into the data group
    write_NXdata(
        dummy_nexus_file,
        [Path("tmp")],
        mock_goniometer.axes_list,
        "images",
        osc_scan,
    )

    write_NXsample(
        dummy_nexus_file,
        mock_goniometer.axes_list,
        "images",
        osc_scan,
    )

    assert f"sample_{test_axis}" in dummy_nexus_file["/entry/sample"]
    assert_array_equal(test_scan_range, dummy_nexus_file[axis_entry][:])
    assert dummy_nexus_file[axis_entry].attrs["depends_on"] == b"."
    assert dummy_nexus_file[axis_entry].attrs["transformation_type"] == b"rotation"
    assert dummy_nexus_file[axis_entry].attrs["units"] == b"deg"
    assert_array_equal(dummy_nexus_file[axis_entry].attrs["vector"][:], [-1, 0, 0])
    assert_array_equal(dummy_nexus_file[axis_entry + "_increment_set"][()], 1)
    # assert_array_equal(dummy_nexus_file[axis_entry + "_increment_set"][:], [1] * 3)
    assert dummy_nexus_file[axis_entry + "_end"][1] == 2


def test_given_reverse_rotation_scan_increment_set_and_axis_end_written_correctly(
    dummy_nexus_file,
):
    test_axis = Axis("phi", ".", TransformationType.ROTATION, (0, 0, -1))
    test_rw_scan = {"phi": np.arange(10, 8, -0.5)}
    test_gonio = Goniometer([test_axis], test_rw_scan)

    # Doing this to write the scan axis data into the data group
    write_NXdata(
        dummy_nexus_file,
        [Path("tmp")],
        test_gonio.axes_list,
        "images",
        test_gonio.scan,
    )

    write_NXsample(
        dummy_nexus_file,
        test_gonio.axes_list,
        "images",
        test_rw_scan,
        sample_depends_on=test_axis.name,
    )

    axis_entry = f"/entry/sample/sample_{test_axis.name}/{test_axis.name}"

    assert_array_equal(dummy_nexus_file[axis_entry][()], [10.0, 9.5, 9.0, 8.5])
    assert_array_equal(dummy_nexus_file[axis_entry + "_increment_set"][()], -0.5)
    assert_array_equal(dummy_nexus_file[axis_entry + "_end"][()], [9.5, 9.0, 8.5, 8.0])


def test_sample_depends_on_written_correctly_in_NXsample(
    dummy_nexus_file, mock_goniometer
):
    test_axis = "omega"
    test_scan_range = [0, 1, 2]
    osc_scan = {test_axis: test_scan_range}

    # Doing this to write the scan axis data into the data group
    write_NXdata(
        dummy_nexus_file,
        [Path("tmp")],
        mock_goniometer.axes_list,
        "images",
        mock_goniometer.scan,
    )

    write_NXsample(
        dummy_nexus_file,
        mock_goniometer.axes_list,
        "images",
        osc_scan,
        sample_depends_on=test_axis,
    )

    assert "depends_on" in dummy_nexus_file["/entry/sample"]
    assert (
        dummy_nexus_file["/entry/sample/depends_on"][()]
        == b"/entry/sample/transformations/omega"
    )


def test_sample_depends_on_written_correctly_in_NXsample_when_value_not_passed(
    dummy_nexus_file,
    mock_goniometer,
):
    test_axis = "omega"
    test_scan_range = [0, 1, 2]
    osc_scan = {test_axis: test_scan_range}

    test_depends = f"/entry/sample/transformations/{mock_goniometer.axes_list[-1].name}"

    # Doing this to write the scan axis data into the data group
    write_NXdata(
        dummy_nexus_file,
        [Path("tmp")],
        mock_goniometer.axes_list,
        "images",
        mock_goniometer.scan,
    )

    write_NXsample(
        dummy_nexus_file,
        mock_goniometer.axes_list,
        "images",
        osc_scan,
    )

    assert "depends_on" in dummy_nexus_file["/entry/sample"]
    assert dummy_nexus_file["/entry/sample/depends_on"][()] == test_depends.encode()


def test_sample_details_in_NXsample(dummy_nexus_file, mock_goniometer):

    test_details = {"name": b"test_sample", "temperature": "25C"}
    test_axis = "omega"
    test_scan_range = [0, 1, 2]
    osc_scan = {test_axis: test_scan_range}

    # Doing this to write the scan axis data into the data group
    write_NXdata(
        dummy_nexus_file,
        [Path("tmp")],
        mock_goniometer.axes_list,
        "images",
        mock_goniometer.scan,
    )

    write_NXsample(
        dummy_nexus_file,
        mock_goniometer.axes_list,
        "images",
        osc_scan,
        sample_details=test_details,
    )

    assert "name" in dummy_nexus_file["/entry/sample"]
    assert "temperature" in dummy_nexus_file["/entry/sample"]
    assert dummy_nexus_file["/entry/sample/name"][()] == test_details["name"]
    assert (
        dummy_nexus_file["/entry/sample/temperature"][()]
        == test_details["temperature"].encode()
    )


def test_given_module_offset_of_1_when_write_NXdetector_module_then_fast_and_slow_axis_depends_on_module_offset(
    dummy_nexus_file,
):
    test_module["module_offset"] = "1"
    write_NXdetector_module(dummy_nexus_file, test_module, [10, 10], [0.1, 0.1], [0, 0])

    module_nexus_path = "/entry/instrument/detector/module/"
    for axis in ["slow_pixel_direction", "fast_pixel_direction"]:
        assert len(dummy_nexus_file[module_nexus_path + axis].attrs.keys()) == 6
        assert (
            dummy_nexus_file[module_nexus_path + axis].attrs["depends_on"]
            == b"/entry/instrument/detector/module/module_offset"
        )
    assert dummy_nexus_file[module_nexus_path + "data_size"].dtype == np.uint16


def test_write_NXdatetime_from_ISO8601str(dummy_nexus_file):
    # Check that ISO8601 strings are accepted and written to file
    entry_path = "/entry/"
    timestamp_str = "2022-03-31T10:30:42Z"
    write_NXdatetime(dummy_nexus_file, timestamp_str, "start_time")

    assert "start_time" in dummy_nexus_file[entry_path].keys()


def test_write_NXdatetime_from_not_ISO8601str(dummy_nexus_file):
    entry_path = "/entry/"
    timestamp_str = datetime.now().strftime("%a %b %d %Y %H:%M:%S")

    write_NXdatetime(dummy_nexus_file, timestamp_str, "start_time")

    assert "start_time" in dummy_nexus_file[entry_path].keys()
    val = dummy_nexus_file[entry_path + "start_time"][()].decode()
    assert val.endswith("Z")


def test_write_NXdatetime_from_datetime(dummy_nexus_file):
    # Check that datetime objects are accepted and written to file
    entry_path = "/entry/"
    end_ts = datetime.fromisoformat("2022-03-31T16:30:32")
    write_NXdatetime(dummy_nexus_file, end_ts, "end_time")

    assert "end_time" in dummy_nexus_file[entry_path].keys()
    val = dummy_nexus_file[entry_path + "end_time"][()].decode()
    assert val.endswith("Z")


def test_write_NXdatetime_writes_nothing_if_wrong_dset_requested(dummy_nexus_file):
    entry_path = "/entry/"
    write_NXdatetime(dummy_nexus_file, "", "no_time")

    assert "no_time" not in dummy_nexus_file[entry_path].keys()


def test_NXdatetime_exits_before_writing_if_timestamp_is_None(dummy_nexus_file):
    entry_path = "/entry/"
    write_NXdatetime(dummy_nexus_file, None, "end_time_estimated")

    assert entry_path + "end_time_estimated" not in dummy_nexus_file.keys()


def test_write_NXnote_in_given_location(dummy_nexus_file):
    loc_path = "/entry/source/pump_probe/"
    info = {"pump_status": True, "pump_exp": 0.001}
    write_NXnote(dummy_nexus_file, loc_path, info)

    assert "pump_status" in dummy_nexus_file[loc_path].keys()
    assert "pump_exp" in dummy_nexus_file[loc_path].keys()
    assert_array_equal(dummy_nexus_file[loc_path + "pump_exp"][()], 0.001)


def test_write_NXcoordinate_system_set(dummy_nexus_file):
    bases = {
        "x": Axis("x", ".", TransformationType.TRANSLATION, (0, 0, 1)),
        "y": Axis("y", "x", TransformationType.TRANSLATION, (0, 0, 0)),
        "z": Axis("z", "y", TransformationType.TRANSLATION, (-1, 0, 0)),
    }

    write_NXcoordinate_system_set(
        dummy_nexus_file, "new_coord_system", bases, (1, 1, 0)
    )

    assert "coordinate_system_set" in dummy_nexus_file["/entry/"].keys()

    loc = "/entry/coordinate_system_set/transformations/"
    assert dummy_nexus_file[loc + "depends_on"][()] == b"."
    assert "x" in dummy_nexus_file[loc].keys()
    assert "y" in dummy_nexus_file[loc].keys()
    assert "z" in dummy_nexus_file[loc].keys()
    assert_array_equal(dummy_nexus_file[loc + "origin"][()], (1, 1, 0))
    assert_array_equal(dummy_nexus_file[loc + "x"][()], 1)
    assert_array_equal(dummy_nexus_file[loc + "y"].attrs["vector"], [0, 0, 0])


def test_write_NXcollection_for_images(dummy_nexus_file):
    nxdet = dummy_nexus_file.require_group("/entry/instrument/detector")
    test_detector_spec = EigerDetector("eiger", [512, 1028], "Si", 10000, -1)
    test_detector_spec.constants["software_version"] = "0.0.0"

    write_NXcollection(nxdet, test_detector_spec, "images", 10)

    spec = "/entry/instrument/detector/detectorSpecific/"
    assert dummy_nexus_file[spec + "software_version"][()] == b"0.0.0"
    assert_array_equal(dummy_nexus_file[spec + "nimages"][()], 10)
    assert_array_equal(
        dummy_nexus_file[spec + "x_pixels"][()], test_detector_spec.image_size[1]
    )
    assert_array_equal(
        dummy_nexus_file[spec + "y_pixels"][()], test_detector_spec.image_size[0]
    )


def test_write_NXcollection_for_events(dummy_nexus_file):
    nxdet = dummy_nexus_file.require_group("/entry/instrument/detector")
    test_detector_spec = TristanDetector("Tristan 1M", [515, 2069])
    test_detector_spec.constants["software_version"] = "0.0.0"

    write_NXcollection(nxdet, test_detector_spec, ("events", None))

    spec = "/entry/instrument/detector/detectorSpecific/"
    assert dummy_nexus_file[spec + "software_version"][()] == b"0.0.0"
    assert_array_equal(
        dummy_nexus_file[spec + "timeslice_rollover_bits"][()],
        test_detector_spec.constants["timeslice_rollover"],
    )
    assert_array_equal(dummy_nexus_file[spec + "detector_tick"][()], 1562.5)
    assert dummy_nexus_file[spec + "detector_tick"].attrs["units"] == b"ps"
    assert dummy_nexus_file[spec + "detector_frequency"].attrs["units"] == b"Hz"


def test_write_NXdetector_for_eiger_images_without_meta_file(
    dummy_nexus_file, mock_eiger
):
    det = "/entry/instrument/detector/"

    write_NXdetector(
        dummy_nexus_file,
        mock_eiger,
        100,
    )

    # Check some general things
    params = mock_eiger.detector_params
    assert dummy_nexus_file[det + "description"][()] == params.description.encode()
    assert dummy_nexus_file[det + "type"][()] == params.detector_type.encode()

    assert_array_equal(
        dummy_nexus_file[det + "beam_center_x"][()], mock_eiger.beam_center[0]
    )
    assert dummy_nexus_file[det + "beam_center_y"].attrs["units"] == b"pixels"
    assert_array_equal(dummy_nexus_file[det + "y_pixel_size"], 7.5e-05)
    assert dummy_nexus_file[det + "x_pixel_size"].attrs["units"] == b"m"

    # Check no mask
    assert "pixel_mask" not in list(dummy_nexus_file[det].keys())

    # Check detector axis and distance
    tr = det + "transformations/"
    assert "detector_z" in list(dummy_nexus_file[tr].keys())
    axis_entry = tr + "detector_z/det_z"
    assert_array_equal(
        mock_eiger.detector_axes[0].start_pos, dummy_nexus_file[axis_entry][()]
    )
    assert dummy_nexus_file[axis_entry].attrs["depends_on"] == b"."
    assert dummy_nexus_file[axis_entry].attrs["transformation_type"] == b"translation"
    assert dummy_nexus_file[axis_entry].attrs["units"] == b"mm"
    assert_array_equal(
        dummy_nexus_file[axis_entry].attrs["vector"][()],
        mock_eiger.detector_axes[0].vector,
    )

    # Check that distance is in meters instead of mm
    assert_array_equal(
        dummy_nexus_file[det + "distance"], mock_eiger.detector_axes[0].start_pos / 1000
    )
    assert dummy_nexus_file[det + "distance"].attrs["units"] == b"m"

    # Check that detector_z has also been written in /detector
    assert "detector_z" in list(dummy_nexus_file[det].keys())


@patch("nexgen.nxs_write.NXclassWriters.write_NXcollection")
def test_write_NXdetector_for_eiger_images_with_meta_file(
    mock_nxcoll_writer,
    dummy_nexus_file,
    dummy_eiger_meta_file,
    mock_eiger,
):
    det = "/entry/instrument/detector/"

    write_NXdetector(
        dummy_nexus_file,
        mock_eiger,
        90,
        Path(dummy_eiger_meta_file.name),
    )
    assert "pixel_mask" in list(dummy_nexus_file[det].keys())
    assert "pixel_mask_applied" in list(dummy_nexus_file[det].keys())
    assert "flatfield" in list(dummy_nexus_file[det].keys())
    assert "bit_depth_readout" in list(dummy_nexus_file[det].keys())

    mock_nxcoll_writer.assert_called_once()


def test_write_NXinstrument(dummy_nexus_file, mock_source, mock_beam, mock_attenuator):
    instr = "/entry/instrument/"

    write_NXinstrument(
        dummy_nexus_file,
        mock_beam,
        mock_attenuator,
        mock_source,
    )

    assert "attenuator" in list(dummy_nexus_file[instr].keys())
    assert "beam" in list(dummy_nexus_file[instr].keys())
    assert dummy_nexus_file[instr + "name"][()] == b"DIAMOND BEAMLINE I03"
    assert dummy_nexus_file[instr + "name"].attrs["short_name"] == b"DLS I03"


def test_write_NXinstrument_sets_correct_instrument_name(
    dummy_nexus_file, mock_attenuator, mock_beam
):
    instr = "/entry/instrument/"

    mock_beam.wavelength = 0.001
    mock_attenuator.transmission = 1.0
    mock_electron_source = Source(
        "eBic",
        Facility(
            "Diamond Light Source", "DLS", "Electron Source", "DIAMOND MICROSCOPE"
        ),
        probe="electron",
    )

    write_NXinstrument(
        dummy_nexus_file,
        mock_beam,
        mock_attenuator,
        mock_electron_source,
        reset_instrument_name=True,
    )

    assert dummy_nexus_file[instr + "name"][()] == b"DIAMOND MICROSCOPE eBic"
    assert dummy_nexus_file[instr + "name"].attrs["short_name"] == b"DLS eBic"



import pytest

from nexgen.nxs_utils.Source import Attenuator, Beam, Facility, Source

test_beam = Beam(0.6)

test_beamline_source = Source("I03")

test_ed_source = Source(
    "m12",
    Facility("Diamond Light Source", "DLS", "Electron Source", "DIAMOND MICROSCOPE"),
    probe="electrons",
)


def test_beam_default_to_dict():
    beamdict = test_beam.to_dict()
    assert beamdict["wavelength"] == 0.6
    assert beamdict["flux"] is None


def test_beam_with_added_flux():
    test_beam.flux = 10
    assert test_beam.wavelength == 0.6
    assert test_beam.to_dict()["flux"] == 10


def test_attenuator_raises_error_if_nothing_is_passed():
    with pytest.raises(TypeError):
        Attenuator()


def test_attenuator():
    assert Attenuator(None).transmission is None
    assert Attenuator(10).transmission == 10


def test_source():
    assert test_beamline_source.beamline == "I03"
    assert test_beamline_source.name == "Diamond Light Source"
    assert test_beamline_source.probe is None

    assert test_beamline_source.set_instrument_name() == "DIAMOND BEAMLINE I03"


def test_dource_to_dict():
    sourcedict = test_beamline_source._generate_source_dict()
    assert "beamline_name" in list(sourcedict.keys())
    assert sourcedict["beamline_name"] == test_beamline_source.beamline
    assert "probe" not in list(sourcedict.keys())


def test_source_for_ed():
    assert test_ed_source.beamline == "m12"
    assert test_ed_source.facility_type == "Electron Source"
    assert test_ed_source.probe == "electrons"

    assert test_ed_source.set_instrument_name() == "DIAMOND MICROSCOPE m12"

    assert test_ed_source._generate_source_dict()["facility_id"] == "DIAMOND MICROSCOPE"



"""
Writer functions for different groups of a NeXus file.
"""
from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, get_args

import h5py  # isort: skip
import numpy as np
from numpy.typing import ArrayLike

from ..nxs_utils import (
    Attenuator,
    Axis,
    Beam,
    Detector,
    DetectorType,
    EigerDetector,
    Source,
)
from ..utils import (
    MAX_SUFFIX_DIGITS,
    get_iso_timestamp,
    units_of_length,
    units_of_time,
    ureg,
)
from .write_utils import (
    TSdset,
    calculate_origin,
    create_attributes,
    mask_and_flatfield_writer,
    mask_and_flatfield_writer_for_event_data,
    set_dependency,
)

NXclass_logger = logging.getLogger("nexgen.NXclass_writers")
NXclass_logger.setLevel(logging.DEBUG)


# NXentry writer
def write_NXentry(nxsfile: h5py.File, definition: str = "NXmx") -> h5py.Group:
    """
    Write NXentry group at top level of the NeXus file.
    Also, write the application definition NXmx.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        definition (str, optional): Application definition for NeXus file. Defaults to "NXmx".

    Returns:
        nxentry (h5py.Group): NXentry group.
    """
    # Set default attribute
    nxsfile.attrs["default"] = "entry"

    # Start writing the NeXus tree with NXentry at the top level
    nxentry = nxsfile.require_group("entry")
    create_attributes(
        nxentry, ("NX_class", "default", "version"), ("NXentry", "data", "1.0")
    )

    # Application definition: /entry/definition
    nxentry.create_dataset("definition", data=np.string_(definition))
    return nxentry


# NXdata writer
def write_NXdata(
    nxsfile: h5py.File,
    datafiles: List[Path],
    goniometer_axes: List[Axis],
    data_type: str,
    osc_scan: Dict[str, ArrayLike],
    transl_scan: Dict[str, ArrayLike] = None,
    entry_key: str = "data",
):
    """
    Write NXdata group at /entry/data.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        datafiles (List[Path]): List of Path objects pointing to HDF5 data files.
        goniometer_axes (List[Axis]): List of goniometer axes.
        data_type (str): Images or events.
        osc_scan (Dict[str, ArrayLike]): Rotation scan. If writing events, this is just a (start, end) tuple.
        transl_scan (Dict[str, ArrayLike], optional): Scan along the xy axes at sample. Defaults to None.
        entry_key (str): Entry key to create the external links to the data files. Defaults to data.

    Raises:
        OSError: If no data is passed.
        ValueError: If the data typs is neither "images" nor "events".
    """
    NXclass_logger.info("Start writing NXdata.")
    # Check that a valid datafile_list has been passed.
    if len(datafiles) == 0:
        raise OSError(
            "No HDF5 data filenames have been found. Please pass at least one."
        )

    # This assumes that a rotation scan is always passed
    osc_axis, osc_range = list(osc_scan.items())[0]

    # Create NXdata group, unless it already exists, in which case just open it.
    nxdata = nxsfile.require_group("/entry/data")
    create_attributes(
        nxdata,
        ("NX_class", "axes", "signal", osc_axis + "_indices"),
        (
            "NXdata",
            osc_axis,
            "data",
            [
                0,
            ],
        ),
    )

    # If mode is images, link to blank image data. Else go to events.
    if data_type == "images":
        tmp_name = f"data_%0{MAX_SUFFIX_DIGITS}d"
        if datafiles[0].parent != Path(nxsfile.filename).parent:
            # This is needed in case the new NeXus file is to be written in a different directory from the data, eg. processing/
            for n, filename in enumerate(datafiles):
                nxdata[tmp_name % (n + 1)] = h5py.ExternalLink(filename, entry_key)
        else:
            for n, filename in enumerate(datafiles):
                nxdata[tmp_name % (n + 1)] = h5py.ExternalLink(filename.name, entry_key)
    elif data_type == "events":
        if len(datafiles) == 1 and "meta" in datafiles[0].as_posix():
            meta = datafiles
        else:
            # Look for meta file to avoid linking to up to 100 files
            tbr = datafiles[0].stem.split("_")[-1]
            mf = datafiles[0].stem.replace(tbr, "meta") + datafiles[0].suffix
            meta = [f for f in datafiles[0].parent.iterdir() if mf in f.as_posix()]
        # If metafile is not found, link to the data files
        if len(meta) == 0:
            for filename in datafiles:
                nxdata[filename.stem] = h5py.ExternalLink(filename.name, "/")
        else:
            nxdata["meta_file"] = h5py.ExternalLink(meta[0].name, "/")
    else:
        raise ValueError(
            "Unknown data type. Please pass one value for data_type from : [images, events]"
        )

    # Write rotation axis dataset
    ax = nxdata.create_dataset(osc_axis, data=osc_range)
    idx = [n for n, ax in enumerate(goniometer_axes) if ax.name == osc_axis][0]
    dep = set_dependency(
        goniometer_axes[idx].depends, path="/entry/sample/transformations/"
    )

    # Write attributes for axis
    create_attributes(
        ax,
        ("depends_on", "transformation_type", "units", "vector"),
        (
            dep,
            goniometer_axes[idx].transformation_type,
            goniometer_axes[idx].units,
            goniometer_axes[idx].vector,
        ),
    )

    # If present, add linear/grid scan details
    if transl_scan:
        for k, v in transl_scan.items():
            ax_dset = nxdata.create_dataset(k, data=v)
            ax_idx = [n for n, ax in enumerate(goniometer_axes) if ax.name == k][0]
            ax_dep = set_dependency(
                goniometer_axes[ax_idx].depends, path="/entry/sample/transformations/"
            )
            create_attributes(
                ax_dset,
                ("depends_on", "transformation_type", "units", "vector"),
                (
                    ax_dep,
                    goniometer_axes[ax_idx].transformation_type,
                    goniometer_axes[ax_idx].units,
                    goniometer_axes[ax_idx].vector,
                ),
            )


# NXsample
def write_NXsample(
    nxsfile: h5py.File,
    goniometer_axes: List[Axis],
    data_type: str,
    osc_scan: Dict[str, ArrayLike],
    transl_scan: Dict[str, ArrayLike] = None,
    sample_depends_on: str = None,
    sample_details: Dict[str, Any] = None,
):
    """
    Write NXsample group at /entry/sample.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        goniometer_axes (List[Axis]): List of goniometer axes.
        data_type (str): Images or events.
        osc_scan (Dict[str, ArrayLike]): Rotation scan. If writing events, this is just a (start, end) tuple.
        transl_scan (Dict[str, ArrayLike], optional): Scan along the xy axes at sample. Defaults to None.
        sample_depends_on (str, optional): Axis on which the sample depends on. If absent, the depends_on field will be set to the last axis listed in the goniometer. Defaults to None.
        sample_details (Dict[str, Any], optional): General information about the sample, eg. name, temperature.
    """
    NXclass_logger.info("Start writing NXsample and NXtransformations.")
    # Create NXsample group, unless it already exists, in which case just open it.
    nxsample = nxsfile.require_group("/entry/sample")
    create_attributes(
        nxsample,
        ("NX_class",),
        ("NXsample",),
    )

    # Create NXtransformations group: /entry/sample/transformations
    nxtransformations = nxsample.require_group("transformations")
    create_attributes(
        nxtransformations,
        ("NX_class",),
        ("NXtransformations",),
    )

    # Get rotation details
    osc_axis, osc_range = list(osc_scan.items())[0]

    # Save sample depends_on
    if sample_depends_on:
        nxsample.create_dataset(
            "depends_on",
            data=set_dependency(sample_depends_on, path=nxtransformations.name),
        )
    else:
        nxsample.create_dataset(
            "depends_on",
            data=set_dependency(goniometer_axes[-1].name, path=nxtransformations.name),
        )

    # Get xy details if passed
    scan_axes = []
    if transl_scan:
        for k in transl_scan.keys():
            scan_axes.append(k)

    # Create sample_{axisname} groups
    for idx, ax in enumerate(goniometer_axes):
        axis_name = ax.name
        grp_name = (
            f"sample_{axis_name[-1]}" if "sam_" in axis_name else f"sample_{axis_name}"
        )
        nxsample_ax = nxsample.create_group(grp_name)
        create_attributes(nxsample_ax, ("NX_class",), ("NXpositioner",))
        if axis_name == osc_axis:
            # If we're dealing with the scan axis
            if (
                "data" in nxsfile["/entry"].keys()
                and axis_name in nxsfile["/entry/data"].keys()
            ):
                nxsample_ax[axis_name] = nxsfile[nxsfile["/entry/data"][axis_name].name]
                nxtransformations[axis_name] = nxsfile[
                    nxsfile["/entry/data"][axis_name].name
                ]
            else:
                nxax = nxsample_ax.create_dataset(axis_name, data=osc_range)
                _dep = set_dependency(
                    goniometer_axes[idx].depends, path="/entry/sample/transformations/"
                )
                create_attributes(
                    nxax,
                    ("depends_on", "transformation_type", "units", "vector"),
                    (
                        _dep,
                        goniometer_axes[idx].transformation_type,
                        goniometer_axes[idx].units,
                        goniometer_axes[idx].vector,
                    ),
                )
                nxtransformations[axis_name] = nxsfile[nxax.name]
            # Write {axisname}_increment_set and {axis_name}_end datasets
            if data_type == "images":
                increment_set = np.repeat(
                    goniometer_axes[idx].increment, len(osc_range)
                )
                nxsample_ax.create_dataset(
                    axis_name + "_increment_set",
                    data=goniometer_axes[idx].increment,
                )  # increment_set
                nxsample_ax.create_dataset(
                    axis_name + "_end", data=osc_range + increment_set
                )
        elif axis_name in scan_axes:
            # For translations
            if (
                "data" in nxsfile["/entry"].keys()
                and axis_name in nxsfile["/entry/data"].keys()
            ):
                nxsample_ax[axis_name] = nxsfile[nxsfile["/entry/data"][axis_name].name]
                nxtransformations[axis_name] = nxsfile[
                    nxsfile["/entry/data"][axis_name].name
                ]
            else:
                nxax = nxsample_ax.create_dataset(
                    axis_name, data=transl_scan[axis_name]
                )
                _dep = set_dependency(
                    goniometer_axes[idx].depends, path="/entry/sample/transformations/"
                )
                create_attributes(
                    nxax,
                    ("depends_on", "transformation_type", "units", "vector"),
                    (
                        _dep,
                        goniometer_axes[idx].transformation_type,
                        goniometer_axes[idx].units,
                        goniometer_axes[idx].vector,
                    ),
                )
                nxtransformations[axis_name] = nxsfile[nxax.name]
        else:
            # For all other axes
            nxax = nxsample_ax.create_dataset(
                axis_name, data=np.array([goniometer_axes[idx].start_pos])
            )
            _dep = set_dependency(
                goniometer_axes[idx].depends, path="/entry/sample/transformations/"
            )
            create_attributes(
                nxax,
                ("depends_on", "transformation_type", "units", "vector"),
                (
                    _dep,
                    goniometer_axes[idx].transformation_type,
                    goniometer_axes[idx].units,
                    goniometer_axes[idx].vector,
                ),
            )
            nxtransformations[axis_name] = nxsfile[nxax.name]

    # Look for nxbeam in file, if it's there make link
    try:
        nxsample["beam"] = nxsfile["/entry/instrument/beam"]
    except KeyError:
        NXclass_logger.debug(
            "No NXbeam group found elsewhere in the NeXus file." "No link written."
        )

    if sample_details:
        for k, v in sample_details.items():
            if isinstance(v, str):
                v = np.string_(v)
            nxsample.create_dataset(k, data=v)


# NXinstrument
def write_NXinstrument(
    nxsfile: h5py.File,
    beam: Beam,
    attenuator: Attenuator,
    source: Source,
    reset_instrument_name: bool = False,
):
    """
    Write NXinstrument group at /entry/instrument.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        beam (Dict):Dictionary with beam information, mainly wavelength and flux.
        attenuator (Dict): Dictionary containing transmission.
        source (Source): Source definition, containing the facility information.
        reset_instrument_name (bool, optional): If True, a string with the name of the \
            instrument used. Otherwise, it will be set to 'DIAMOND BEAMLINE Ixx'. Defaults to False.
    """
    NXclass_logger.info("Start writing NXinstrument.")
    # Create NXinstrument group, unless it already exists, in which case just open it.
    nxinstrument = nxsfile.require_group("/entry/instrument")
    create_attributes(
        nxinstrument,
        ("NX_class",),
        ("NXinstrument",),
    )

    # Write /name field and relative attribute
    NXclass_logger.info(f"{source.short_name} {source.beamline}")
    name_str = (
        source.set_instrument_name()
        if reset_instrument_name
        else f"DIAMOND BEAMLINE {source.beamline}"
    )
    nxinstrument.create_dataset("name", data=np.string_(name_str))
    create_attributes(
        nxinstrument["name"],
        ("short_name",),
        (f"{source.short_name} {source.beamline}",),
    )

    NXclass_logger.info("Write NXattenuator and NXbeam.")
    # Write NXattenuator group: entry/instrument/attenuator
    nxatt = nxinstrument.require_group("attenuator")
    create_attributes(nxatt, ("NX_class",), ("NXattenuator",))
    if attenuator.transmission:
        nxatt.create_dataset(
            "attenuator_transmission",
            data=attenuator.transmission,
        )

    # Write NXbeam group: entry/instrument/beam
    nxbeam = nxinstrument.require_group("beam")
    create_attributes(nxbeam, ("NX_class",), ("NXbeam",))
    wl = nxbeam.create_dataset("incident_wavelength", data=beam.wavelength)
    create_attributes(wl, ("units",), ("angstrom",))
    if beam.flux:
        flux = nxbeam.create_dataset("total_flux", data=beam.flux)
        create_attributes(flux, ("units"), ("Hz",))


# NXsource
def write_NXsource(nxsfile: h5py.File, source: Source):
    """
    Write NXsource group /in entry/source.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        source (Source): Source definition, containing the facility information.
    """
    NXclass_logger.info("Start writing NXsource.")
    nxsource = nxsfile.require_group("/entry/source")
    create_attributes(
        nxsource,
        ("NX_class",),
        ("NXsource",),
    )

    nxsource.create_dataset("name", data=np.string_(source.name))
    create_attributes(nxsource["name"], ("short_name",), (source.short_name,))
    nxsource.create_dataset("type", data=np.string_(source.facility_type))
    if source.probe:
        nxsource.create_dataset("probe", data=np.string_(source.probe))


# NXdetector writer
def write_NXdetector(
    nxsfile: h5py.File,
    detector: Detector,
    num_images: int = None,
    meta: Path = None,
):
    """
    Write_NXdetector group at /entry/instrument/detector.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        detector (Detector): Detector definition.
        num_images (int, optional): Total number of images in collections. Defaults to None
        meta (Path, optional): Path to _meta.h5 file. Defaults to None.
    """
    NXclass_logger.info("Start writing NXdetector.")
    # Create NXdetector group, unless it already exists, in which case just open it.
    nxdetector = nxsfile.require_group("/entry/instrument/detector")
    create_attributes(
        nxdetector,
        ("NX_class",),
        ("NXdetector",),
    )

    # Detector description
    nxdetector.create_dataset(
        "description", data=np.string_(detector.detector_params.description)
    )
    nxdetector.create_dataset(
        "type", data=np.string_(detector.detector_params.detector_type)
    )

    collection_mode = detector.get_detector_mode()

    # If there is a meta file, a lot of information will be linked instead of copied
    if isinstance(detector.detector_params, EigerDetector):
        if meta:
            NXclass_logger.info(f"Found metadata in {meta.as_posix()} file.")
            meta_link = (
                meta.name
                if meta.parent == Path(nxsfile.filename).parent
                else meta.as_posix()
            )
            for k, v in detector.detector_params.constants.items():
                if k == "software_version":
                    # Software version should go in detectorSpecific (NXcollection)
                    break
                nxdetector[k] = h5py.ExternalLink(meta_link, v)
        else:
            NXclass_logger.warning(
                """
                Meta file for Eiger detector hasn't been specified.
                No links will be written. Pixel mask and flatfield information missing.
                """
            )
    else:
        # If it's an eiger mask and flatfield will be in the links along with other info.
        # If it isn't, the mask and flatfield info still needs to go in
        NXclass_logger.info("Gathering and writing pixel_mask and flatfield to file.")
        pixel_mask_file = detector.detector_params.constants["pixel_mask"]
        flatfield_file = detector.detector_params.constants["flatfield"]
        if collection_mode == "events":
            wd = Path(nxsfile.filename).parent
            # Bad pixel mask
            mask_and_flatfield_writer_for_event_data(
                nxdetector,
                "pixel_mask",
                pixel_mask_file,
                detector.detector_params.constants["pixel_mask_applied"],
                wd,
                detector.detector_params.description.lower(),
            )
            # Flatfield
            mask_and_flatfield_writer_for_event_data(
                nxdetector,
                "flatfield",
                flatfield_file,
                detector.detector_params.constants["flatfield_applied"],
                wd,
                detector.detector_params.description.lower(),
            )
        else:
            # Flatfield
            mask_and_flatfield_writer(
                nxdetector,
                "flatfield",
                detector.detector_params.constants["flatfield"],
                detector.detector_params.constants["flatfield_applied"],
            )
            # Bad pixel mask
            mask_and_flatfield_writer(
                nxdetector,
                "pixel_mask",
                detector.detector_params.constants["pixel_mask"],
                detector.detector_params.constants["pixel_mask_applied"],
            )

    # Beam center
    beam_center_x = nxdetector.create_dataset(
        "beam_center_x", data=detector.beam_center[0]
    )
    create_attributes(beam_center_x, ("units",), ("pixels",))
    beam_center_y = nxdetector.create_dataset(
        "beam_center_y", data=detector.beam_center[1]
    )
    create_attributes(beam_center_y, ("units",), ("pixels",))

    # Pixel size in m
    x_pix = units_of_length(detector.detector_params.pixel_size[0], True)
    x_pix_size = nxdetector.create_dataset("x_pixel_size", data=x_pix.magnitude)
    create_attributes(x_pix_size, ("units",), (format(x_pix.units, "~"),))
    y_pix = units_of_length(detector.detector_params.pixel_size[1], True)
    y_pix_size = nxdetector.create_dataset("y_pixel_size", data=y_pix.magnitude)
    create_attributes(y_pix_size, ("units",), (format(y_pix.units, "~"),))

    # Sensor material, sensor thickness in m
    nxdetector.create_dataset(
        "sensor_material", data=np.string_(detector.detector_params.sensor_material)
    )
    sensor_thickness = units_of_length(detector.detector_params.sensor_thickness, True)
    nxdetector.create_dataset("sensor_thickness", data=sensor_thickness.magnitude)
    create_attributes(
        nxdetector["sensor_thickness"],
        ("units",),
        (format(sensor_thickness.units, "~"),),
    )

    # Count time
    exp_time = units_of_time(detector.exp_time)
    nxdetector.create_dataset("count_time", data=exp_time.magnitude)

    # If detector mode is images write overload and underload
    if collection_mode == "images":
        nxdetector.create_dataset(
            "saturation_value", data=detector.detector_params.overload
        )
        nxdetector.create_dataset(
            "underload_value", data=detector.detector_params.underload
        )

    # Write_NXcollection
    write_NXcollection(
        nxdetector, detector.detector_params, collection_mode, num_images, meta
    )

    # Write NXtransformations: entry/instrument/detector/transformations/detector_z and two_theta
    nxtransformations = nxdetector.require_group("transformations")
    create_attributes(
        nxtransformations,
        ("NX_class",),
        ("NXtransformations",),
    )

    # Create groups for detector_z and any other detector axis (eg. two_theta) if present
    # This assumes that the detector axes are fixed.
    for idx, ax in enumerate(detector.detector_axes):
        if ax.name == "det_z":
            grp_name = "detector_z"
            dist = units_of_length(str(detector.detector_axes[idx].start_pos) + "mm")
        else:
            grp_name = ax.name

        # It shouldn't be too much of an issue but just in case ...
        if detector.detector_axes[idx].depends == "det_z":
            grp_dep = "detector_z"
        else:
            grp_dep = detector.detector_axes[idx].depends
        _dep = set_dependency(
            detector.detector_axes[idx].depends,
            nxtransformations.name + f"/{grp_dep}/",
        )

        nxgrp_ax = nxtransformations.create_group(grp_name)
        create_attributes(nxgrp_ax, ("NX_class",), ("NXpositioner",))
        nxdet_ax = nxgrp_ax.create_dataset(
            ax.name, data=np.array([detector.detector_axes[idx].start_pos])
        )
        create_attributes(
            nxdet_ax,
            ("depends_on", "transformation_type", "units", "vector"),
            (
                _dep,
                detector.detector_axes[idx].transformation_type,
                detector.detector_axes[idx].units,
                detector.detector_axes[idx].vector,
            ),
        )
        if ax.name == detector.detector_axes[-1].name:
            # Detector depends_on
            nxdetector.create_dataset(
                "depends_on",
                data=set_dependency(ax.name, path=nxgrp_ax.name),
            )

    # Write a soft link for detector_z
    if "detector_z" in list(nxtransformations.keys()):
        nxdetector["detector_z"] = nxsfile[
            "/entry/instrument/detector/transformations/detector_z"
        ]

    # Detector distance
    nxdetector.create_dataset("distance", data=dist.to("m").magnitude)
    create_attributes(
        nxdetector["distance"], ("units",), (format(dist.to("m").units, "~"))
    )

    # Check if there are any remaining datasets to be written (usually from the meta file but not always)
    others = [
        "threshold_energy",
        "bit_depth_image",
        "detector_number",
        "detector_readout_time",
        "photon_energy",
    ]
    for dset in others:
        if (
            nxdetector.__contains__(dset) is False
            and dset in detector.detector_params.constants.keys()
        ):
            val = (
                np.string_(detector.detector_params.constants[dset])
                if isinstance(detector.detector_params.constants[dset], str)
                else detector.detector_params.constants[dset]
            )
            if val is not None:
                nxdetector.create_dataset(dset, data=val)


# NXdetector_module writer
def write_NXdetector_module(
    nxsfile: h5py.File,
    module: Dict,
    image_size: List | Tuple,
    pixel_size: List | Tuple,
    beam_center: Optional[List | Tuple] = None,
):
    """
    Write NXdetector_module group at /entry/instrument/detector/module.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        module (Dict): Dictionary containing the detector module information: fast and slow axes, how many modules.
        image_size (List | Tuple): Size of the detector, in pixels, passed in the order (slow, fast) axis.
        pixel_size (List | Tuple): Size of the single pixels in fast and slow direction, in mm.
        beam_center (Optional[List | Tuple], optional): Beam center position, needed only if origin needs to be calculated. Defaults to None.
    """
    NXclass_logger.info("Start writing NXdetector_module.")
    # Create NXdetector_module group, unless it already exists, in which case just open it.
    nxmodule = nxsfile.require_group("/entry/instrument/detector/module")
    create_attributes(
        nxmodule,
        ("NX_class",),
        ("NXdetector_module",),
    )

    nxmodule.create_dataset("data_origin", data=np.array([0, 0]), dtype=np.uint16)
    nxmodule.create_dataset("data_size", data=image_size, dtype=np.uint16)
    nxmodule.create_dataset("data_stride", data=np.array([1, 1]), dtype=np.uint16)

    # Write fast_ and slow_ pixel_direction
    fast_axis = module["fast_axis"]
    slow_axis = module["slow_axis"]

    if "offsets" in module.keys():
        offsets = module["offsets"]
    else:
        offsets = [(0, 0, 0), (0, 0, 0)]

    x_pix = units_of_length(pixel_size[0], True)
    fast_pixel = nxmodule.create_dataset("fast_pixel_direction", data=x_pix.magnitude)
    create_attributes(
        fast_pixel,
        (
            "depends_on",
            "offset",
            "offset_units",
            "transformation_type",
            "units",
            "vector",
        ),
        (
            "/entry/instrument/detector/transformations/detector_z/det_z",
            offsets[0],
            "mm",
            "translation",
            format(x_pix.units, "~"),
            fast_axis,
        ),
    )

    y_pix = units_of_length(pixel_size[1], True)
    slow_pixel = nxmodule.create_dataset("slow_pixel_direction", data=y_pix.magnitude)
    create_attributes(
        slow_pixel,
        (
            "depends_on",
            "offset",
            "offset_units",
            "transformation_type",
            "units",
            "vector",
        ),
        (
            "/entry/instrument/detector/module/fast_pixel_direction",
            offsets[1],
            "mm",
            "translation",
            format(y_pix.units, "~"),
            slow_axis,
        ),
    )

    # If module_offset is set to 1 or 2, calculate accordinlgy and write the field
    if "module_offset" not in module.keys():
        NXclass_logger.warning(
            "Module_offset option wasn't passed."
            "It will be automatically set to '1' and the origin calculated accordingly."
            "To skip this calculation please set module_offset to '0'."
        )
        module["module_offset"] = "1"

    if module["module_offset"] != "0":
        pixel_size_m = [
            x_pix.magnitude,
            y_pix.magnitude,
        ]
        origin, offset_val = calculate_origin(
            beam_center,
            pixel_size_m,
            fast_axis,
            slow_axis,
            mode=module["module_offset"],
        )
        module_offset = nxmodule.create_dataset("module_offset", data=offset_val)
        create_attributes(
            module_offset,
            (
                "depends_on",
                "offset",
                "offset_units",
                "transformation_type",
                "units",
                "vector",
            ),
            (
                "/entry/instrument/detector/transformations/detector_z/det_z",
                [0, 0, 0],
                "mm",
                "translation",
                format(x_pix.units, "~"),
                origin,
            ),
        )

        # Correct dependency tree accordingly
        _path = "/entry/instrument/detector/module/module_offset"
        create_attributes(fast_pixel, ("depends_on",), (_path,))
        create_attributes(slow_pixel, ("depends_on",), (_path,))


# NXCollection writer (detectorSpecific)
def write_NXcollection(
    nxdetector: h5py.Group,
    detector_params: DetectorType,
    collection_mode: str = "images",
    num_images: int = None,
    meta: Path = None,
):
    """
    Write a NXcollection group inside NXdetector as detectorSpecific.

    Args:
        nxdetector (h5py.Group): HDF5 NXdetector group handle.
        detector_params (DetectorType): Parameters specific to the detector in use.
        collection_mode (str, optional): Data type collected by detector. Defaults to "images".
        num_images (int, optional): Total number of images collected. Defaults to None.
        meta (Path, optional): Path to _meta.h5 file. Defaults to None.
    """
    NXclass_logger.info("Start writing detectorSpecific group as NXcollection.")
    # Create detectorSpecific group
    grp = nxdetector.require_group("detectorSpecific")
    grp.create_dataset(
        "x_pixels", data=detector_params.image_size[1], dtype=np.uint16
    )  # fast axis
    grp.create_dataset(
        "y_pixels", data=detector_params.image_size[0], dtype=np.uint16
    )  # slow axis
    # Write these non-spec fields as well because of autoPROC:
    grp.create_dataset(
        "x_pixels_in_detector", data=detector_params.image_size[1], dtype=np.uint16
    )  # fast axis
    grp.create_dataset(
        "y_pixels_in_detector", data=detector_params.image_size[0], dtype=np.uint16
    )  # slow axis
    if collection_mode == "images":
        grp.create_dataset("nimages", data=num_images)
    if "software_version" in list(detector_params.constants.keys()):
        if not detector_params.hasMeta or collection_mode == "events":
            grp.create_dataset(
                "software_version",
                data=np.string_(detector_params.constants["software_version"]),
            )
        elif detector_params.hasMeta and meta:
            grp["software_version"] = h5py.ExternalLink(
                meta.name, detector_params.constants["software_version"]
            )
        else:
            grp.create_dataset(
                "software_version",
                data=np.string_(detector_params.constants["software_version"]),
            )
    if "TRISTAN" in detector_params.description.upper():
        tick = ureg.Quantity(detector_params.constants["detector_tick"])
        grp.create_dataset("detector_tick", data=tick.magnitude)
        grp["detector_tick"].attrs["units"] = np.string_(format(tick.units, "~"))
        freq = ureg.Quantity(detector_params.constants["detector_frequency"])
        grp.create_dataset("detector_frequency", data=freq.magnitude)
        grp["detector_frequency"].attrs["units"] = np.string_(format(freq.units, "~"))
        grp.create_dataset(
            "timeslice_rollover_bits",
            data=detector_params.constants["timeslice_rollover"],
        )


# NXdatetime writer
def write_NXdatetime(
    nxsfile: h5py.File,
    timestamp: datetime | str,
    dset_name: TSdset = "start_time",
):
    """Write NX_DATE_TIME fields under /entry/.
    Required fields for NXmx format: 'start_time' and 'end_time_estimated'.
    Optional field: 'end_time', to be writte only if values accurately observed

    Args:
        nxsfile (h5py.File): Nexus file handle.
        timestamp (datetime | str): Timestamp, either in datetime or as a string.
        dset_name (TSdset, optional): NXdatetime dataset name.\
            Allowed values: ["start_time", "end_time", "end_time_estimated". Defaults to "start_time".
    """
    if timestamp is None:
        NXclass_logger.warning(
            f"Timestamp value is None, {dset_name} won't be written."
        )
        return
    nxentry = nxsfile.require_group("entry")
    dset_opts = get_args(TSdset)
    if dset_name not in dset_opts:
        NXclass_logger.warning(
            f"{dset_name} is not an allowed value for NXdatetime dataset. Please pass one of {dset_opts}."
        )
        return
    if type(timestamp) is datetime:
        timestamp = timestamp.strftime("%Y-%m-%dT%H:%M:%S")
    timestamp = get_iso_timestamp(timestamp)
    nxentry.create_dataset(dset_name, data=np.string_(timestamp))


# NXnote writer
# To be used e.g. as a place to store pump-probe info such as pump delay/width
def write_NXnote(nxsfile: h5py.File, loc: str, info: Dict):
    """
    Write any additional information as a NXnote class in a specified location in the NeXus file.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        loc (str): Location inside the NeXus file to write NXnote group.
        info (Dict): Dictionary of datasets to be written to NXnote.
    """
    NXclass_logger.info("Start writing NXnote.")
    # Create the NXnote group in the specified location
    nxnote = nxsfile.require_group(loc)
    create_attributes(
        nxnote,
        ("NX_class",),
        ("NXnote",),
    )

    # Write datasets
    for k, v in info.items():
        if v:  # Just in case one value is not recorded and set as None
            if isinstance(v, str):
                v = np.string_(v)
            nxnote.create_dataset(k, data=v)
            NXclass_logger.info(f"{k} dataset written in {loc}.")


def write_NXcoordinate_system_set(
    nxsfile: h5py.File,
    convention: str,
    base_vectors: Dict[str, Axis],
    origin: List | Tuple | ArrayLike,
):
    """
    Write a container object to store coordinate system conventions different from mcstas.

    The NXcoordinate_system_set base class is used here to define and store the mappings between mcstas and a different coordinate system relevant to the data.
    It should hold at least one NXtransformations group containing a depends_on field, which specifies whether this coordinate system is the reference ("."),
    as well as the three base vectors and the location of the origin.

    A template for base_vectors:
    base_vectors = {"x": (depends_on, transformation_type, units, vector), ...}

    Args:
        nxsfile (h5py.File): Handle to NeXus file.
        convention (str): Convention decription. Defaults to "ED".
        base_vectors (Dict[str, Axis]): The three base vectors of the coordinate system.
        origin (List | Tuple | np.ndarray): The location of the origin of the coordinate system.
    """
    NXclass_logger.info(
        f"Writing NXcoordinate_system_set to define the coordinate system convention for {convention}."
    )

    #
    nxcoord = nxsfile.require_group("/entry/coordinate_system_set")
    create_attributes(
        nxcoord,
        ("NXclass",),
        ("NXcoordinate_system_set",),
    )

    transf = nxcoord.require_group("transformations")
    create_attributes(
        transf,
        ("NXclass",),
        ("NXtransformations",),
    )

    # Needs at least: 3 base vectors, depends_on ("." probably?), origin
    transf.create_dataset("depends_on", data=np.string_("."))  # To be checked
    transf.create_dataset("origin", data=origin)

    # Base vectors
    NXclass_logger.info(
        "Base vectors: \n"
        f"x: {base_vectors['x'].vector} \n"
        f"y: {base_vectors['y'].vector} \n"
        f"z: {base_vectors['z'].vector} \n"
    )
    idx = 0
    for k, v in base_vectors.items():
        base = transf.create_dataset(k, data=np.array(origin[idx]))
        create_attributes(
            base,
            ("depends_on", "transformation_type", "units", "vector"),
            (
                set_dependency(v.depends, transf.name),
                v.transformation_type,
                v.units,
                v.vector,
            ),
        )
        idx += 1



import pytest

from nexgen.command_line.cli_utils import (
    reframe_arrays,
    set_detector_params,
    split_arrays,
)
from nexgen.nxs_utils.Detector import UnknownDetectorTypeError

test_goniometer_small = {
    "axes": ["alpha", "sam_z"],
    "vectors": [1, 0, 0, 0, 0, 1],
    "offsets": [(0, 0, 0), (0, 0, 0)],
}

test_detector = {"axes": ["det_z"], "vectors": [0, 0, 1]}
test_module = {
    "fast_axis": [1, 0, 0],
    "slow_axis": [0, 1, 0],
}

test_new_coords = {
    "convention": "rotate",
    "origin": (0, 0, 0),
    "x": (".", "", "", [1, 0, 0]),
    "y": (".", "", "", [0, -1, 0]),
    "z": (".", "", "", [0, 0, -1]),
}


def test_reframe_arrays_without_coordinate_conversion():
    reframe_arrays(
        test_goniometer_small,
        test_detector,
        test_module,
    )
    assert test_goniometer_small["vectors"] == [(1, 0, 0), (0, 0, 1)]
    assert test_detector["vectors"] == [(0, 0, 1)]


def test_reframe_arrays_from_imgcif():
    reframe_arrays(
        test_goniometer_small,
        test_detector,
        test_module,
        "imgcif",
    )
    assert test_goniometer_small["vectors"] == [(-1, 0, 0), (0, 0, -1)]
    assert test_detector["vectors"] == [(0, 0, -1)]
    assert test_module["fast_axis"] == (-1, 0, 0)
    assert test_module["slow_axis"] == (0, 1, 0)


def test_reframe_arrays_from_another_coordinate_system():
    test_goniometer_small["vectors"] = [1, 0, 0, 0, 0, 1]
    test_detector["vectors"] = [0, 0, 1]
    test_module["fast_axis"] = [1, 0, 0]
    test_module["slow_axis"] = [0, 1, 0]
    test_module["offsets"] = [0, 0, 0, 0, 0, 0]
    reframe_arrays(
        test_goniometer_small,
        test_detector,
        test_module,
        "rotate",
        test_new_coords,
    )

    assert test_goniometer_small["vectors"] == [(1, 0, 0), (0, 0, -1)]
    assert test_detector["vectors"] == [(0, 0, -1)]
    assert test_module["fast_axis"] == (1, 0, 0)
    assert test_module["slow_axis"] == (0, -1, 0)
    assert test_module["offsets"] == [(0, 0, 0), (0, 0, 0)]


def test_reframe_arrays_fails_if_coordinate_system_ill_defined():
    with pytest.raises(ValueError):
        reframe_arrays(
            test_goniometer_small, test_detector, test_module, "", test_new_coords
        )


def test_reframe_arrays_fails_if_new_coordinate_system_not_defined():
    with pytest.raises(TypeError):
        reframe_arrays(test_goniometer_small, test_detector, test_module, "new")


def test_split_arrays():
    assert split_arrays(["phi"], [1, 0, 0]) == {"phi": (1, 0, 0)}
    two_axes = split_arrays(["omega", "phi"], [1, 0, 0, 0, 1, 0])
    assert two_axes["omega"] == (1, 0, 0) and two_axes["phi"] == (0, 1, 0)
    assert (
        len(split_arrays(["omega", "phi", "chi"], [(1, 0, 0), (0, 1, 0), (0, 0, 1)]))
        == 3
    )


def test_split_arrays_fails_if_wrong_size_arrays():
    with pytest.raises(ValueError):
        split_arrays(["omega", "phi"], [1, 0, 0, 1])


def test_set_detector_params_fails_if_missing_kwargs():
    with pytest.raises(ValueError):
        set_detector_params("tristan", image_size=[1, 2])


def test_set_detector_params_fails_if_neither_tristan_nor_eiger():
    with pytest.raises(UnknownDetectorTypeError):
        set_detector_params("singla")



"""
Create a NeXus file for serial crystallography datasets collected on Eiger detector either on I19-2 or I24 beamlines.
"""
from __future__ import annotations

import logging
from collections import namedtuple
from pathlib import Path

import h5py

from .. import log
from ..nxs_utils import Attenuator, Beam, Detector, EigerDetector, Goniometer, Source
from ..nxs_write.NXmxWriter import NXmxFileWriter
from ..tools.Metafile import DectrisMetafile
from ..tools.MetaReader import define_vds_data_type, update_axes_from_meta
from ..utils import find_in_dict, get_iso_timestamp
from .beamline_utils import PumpProbe, collection_summary_log

# Define logger
logger = logging.getLogger("nexgen.SSX_Eiger")

# Define a namedtuple for collection parameters
ssx_collect = namedtuple(
    "ssx_collect",
    [
        "num_imgs",
        "exposure_time",
        "detector_distance",
        "beam_center",
        "transmission",
        "wavelength",
        "flux",
        "start_time",
        "stop_time",
        "chip_info",
        "chipmap",
    ],
)

ssx_collect.__doc__ = """Serial collection parameters"""


def ssx_eiger_writer(
    visitpath: Path | str,
    filename: str,
    beamline: str,
    num_imgs: int,
    expt_type: str = "fixed-target",
    pump_status: bool = False,
    **ssx_params,
):
    """Gather all collection parameters and write the NeXus file for SSX using Eiger detector.

    Args:
        visitpath (Path | str): Collection directory.
        filename (str): Filename root.
        beamline (str): Beamline on which the experiment is being run.
        num_imgs (int): Total number of images collected.
        expt_type (str, optional): Experiment type, accepted values: extruder,
            fixed-target, 3Dgridscan. Defaults to "fixed-target".
        pump_status (bool, optional): True for pump-probe experiment. Defaults to False.

    Keyword Args:
        exp_time (float): Exposure time, in s.
        det_dist (float): Distance between sample and detector, in mm.
        beam_center (List[float, float]): Beam center position, in pixels.
        transmission (float): Attenuator transmission, in %.
        wavelength (float): Wavelength of incident beam, in A.
        flux (float): Total flux.
        start_time (datetime): Experiment start time.
        stop_time (datetime): Experiment end time.
        chip_info (Dict): For a grid scan, dictionary containing basic chip information.
            At least it should contain: x/y_start, x/y number of blocks and block size, x/y number of steps and number of exposures.
        chipmap (Path | str): Path to the chipmap file corresponding to the experiment,
            or 'fullchip' indicating that the whole chip is being scanned.
        pump_exp (float): Pump exposure time, in s.
        pump_delay (float): Pump delay time, in s.
        osc_axis (str): Oscillation axis. Always omega on I24. If not passed it will default to phi for I19-2.

    Raises:
        ValueError: If an invalid beamline name is passed.
        ValueError: If an invalid experiment type is passed.
    """
    if not find_in_dict("start_time", ssx_params):
        ssx_params["start_time"] = None
    if not find_in_dict("stop_time", ssx_params):
        ssx_params["stop_time"] = None

    SSX = ssx_collect(
        num_imgs=int(num_imgs),
        exposure_time=ssx_params["exp_time"]
        if find_in_dict("exp_time", ssx_params)
        else None,
        detector_distance=ssx_params["det_dist"]
        if find_in_dict("det_dist", ssx_params)
        else None,
        beam_center=ssx_params["beam_center"]
        if find_in_dict("beam_center", ssx_params)
        else (0, 0),
        transmission=ssx_params["transmission"]
        if find_in_dict("transmission", ssx_params)
        else None,
        wavelength=ssx_params["wavelength"]
        if find_in_dict("wavelength", ssx_params)
        else None,
        flux=ssx_params["flux"] if find_in_dict("flux", ssx_params) else None,
        start_time=ssx_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["start_time"]
        else None,
        stop_time=ssx_params["stop_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["stop_time"]
        else None,
        chip_info=ssx_params["chip_info"]
        if find_in_dict("chip_info", ssx_params)
        else None,
        chipmap=ssx_params["chipmap"] if find_in_dict("chipmap", ssx_params) else None,
    )

    if expt_type.lower() not in ["extruder", "fixed-target", "3Dgridscan"]:
        raise ValueError("Unknown experiment type.")

    visitpath = Path(visitpath).expanduser().resolve()

    # Configure logging
    logfile = visitpath / f"{beamline}_EigerSSX_nxs_writer.log"
    log.config(logfile.as_posix())

    logger.info(f"Current collection directory: {visitpath.as_posix()}")
    # Get NeXus filename
    master_file = visitpath / f"{filename}.nxs"
    logger.info("NeXus file will be saved as %s" % master_file.as_posix())

    # Get parameters depending on beamline
    logger.info(f"DLS Beamline: {beamline.upper()}.")
    if "I19" in beamline.upper():
        source = Source("I19-2")
        osc_axis = ssx_params["osc_axis"] if "osc_axis" in ssx_params.keys() else "phi"
        from .I19_2_params import I19_2Eiger as axes_params

        eiger_params = EigerDetector(
            "Eiger 2X 4M",
            (2162, 2068),
            "CdTe",
            50649,
            -1,
        )
    elif "I24" in beamline.upper():
        source = Source("I24")
        osc_axis = "omega"
        from .I24_params import I24Eiger as axes_params

        eiger_params = EigerDetector(
            "Eiger 2X 9M",
            (3262, 3108),
            "CdTe",
            50649,
            -1,
        )
    else:
        raise ValueError(
            "Unknown beamline for SSX collections with Eiger detector."
            "Beamlines currently enabled for the writer: I24 (Eiger 9M), I19-2 (Eiger 4M)."
        )

    # Define what to do based on experiment type
    if expt_type not in ["extruder", "fixed-target", "3Dgridscan"]:
        raise ValueError(
            "Please pass a valid experiment type.\n"
            "Accepted values: extruder, fixed-target, 3Dgridscan."
        )
    logger.info(f"Running {expt_type} collection.")

    # Get pump information
    pump_probe = PumpProbe()
    if pump_status is True:
        # Exposure and delay could also be found in dictionary for grid scan
        logger.info("Pump status is True.")
        pump_probe.pump_status = pump_status
        pump_probe.pump_exposure = (
            ssx_params["pump_exp"] if find_in_dict("pump_exp", ssx_params) else None
        )
        pump_probe.pump_delay = (
            ssx_params["pump_exp"] if find_in_dict("pump_exp", ssx_params) else None
        )

        logger.info(f"Recorded pump exposure time: {pump_probe.pump_exposure}")
        logger.info(f"Recorded pump delay time: {pump_probe.pump_delay}")
        if expt_type == "fixed-target":
            pump_probe.pump_repeat = int(SSX.chip_info["PUMP_REPEAT"][1])

    # Get timestamps in the correct format
    timestamps = (
        get_iso_timestamp(SSX.start_time),
        get_iso_timestamp(SSX.stop_time),
    )

    # Find metafile in directory and get info from it
    try:
        metafile = [
            f for f in visitpath.iterdir() if filename + "_meta" in f.as_posix()
        ][0]
        logger.debug(f"Found {metafile} in directory.")
    except IndexError as err:
        logger.exception(err)
        logger.error(
            "Missing metadata, something might be wrong with this collection."
            "Unable to write NeXus file at this time. Please try using command line tool."
        )
        raise

    # Define Attenuator
    attenuator = Attenuator(SSX.transmission)
    # Define Beam
    wl = SSX.wavelength
    if not wl:
        logger.debug("No wavelength passed, looking for it in the meta file.")

        with h5py.File(metafile, "r", libver="latest", swmr=True) as fh:
            wl = DectrisMetafile(fh).get_wavelength()
    beam = Beam(wl, SSX.flux)

    # Define Goniometer axes
    gonio_axes = axes_params.gonio
    # Define Detector
    det_axes = axes_params.det_axes

    # Update axes starts and get data type from meta file
    with h5py.File(metafile, "r", libver="latest", swmr=True) as fh:
        meta = DectrisMetafile(fh)
        vds_dtype = define_vds_data_type(meta)
        update_axes_from_meta(meta, gonio_axes, osc_axis=osc_axis)
        update_axes_from_meta(meta, det_axes)

    logger.debug(
        "Goniometer and detector axes have ben updated with values from the meta file."
    )

    # Sanity check on det_z vs SSX.det_dist
    logger.debug("Sanity check on detector distance.")
    det_z_idx = [n for n, ax in enumerate(det_axes) if ax.name == "det_z"][0]
    if SSX.detector_distance and SSX.detector_distance != det_axes[det_z_idx].start_pos:
        logger.debug(
            "Detector distance value in meta file did not match with the one passed by the user.\n"
            f"Passed value: {SSX.detector_distance}; Value stored in meta file: {det_axes[det_z_idx].start_pos}.\n"
            "Value will be overwritten with the passed one."
        )
        det_axes[det_z_idx].start_pos = SSX.detector_distance

    # Define detector
    detector = Detector(
        eiger_params,
        det_axes,
        SSX.beam_center,
        SSX.exposure_time,
        [axes_params.fast_axis, axes_params.slow_axis],
    )

    tot_num_imgs = SSX.num_imgs

    # Run experiment type
    if expt_type == "extruder":
        from .SSX_expt import run_extruder

        gonio_axes, SCAN, pump_info = run_extruder(
            gonio_axes,
            tot_num_imgs,
            pump_probe,
            osc_axis,
        )
    elif expt_type == "fixed-target":
        from .SSX_expt import run_fixed_target

        # Define chipmap if needed
        chipmapfile = (
            "fullchip"
            if SSX.chipmap is None
            else Path(SSX.chipmap).expanduser().resolve()
        )

        SCAN, pump_info = run_fixed_target(
            gonio_axes,
            SSX.chip_info,
            chipmapfile,
            pump_probe,
            ["sam_y", "sam_x"],
        )

        # Sanity check that things make sense
        if SSX.num_imgs != len(SCAN["sam_x"]):
            logger.warning(
                f"The total number of scan points is {len(SCAN['sam_x'])}, which does not match the total number of images passed as input {SSX.num_imgs}."
            )
            logger.warning(
                "Reset SSX.num_imgs to number of scan points for vds creation."
            )
            tot_num_imgs = len(SCAN["sam_x"])
    else:
        print("3D")
        SCAN = {}  # tboth here here
        pump_info = pump_probe.to_dict()

    # Define goniometer only once the full scan has been calculated.
    goniometer = Goniometer(gonio_axes, SCAN)

    # Log a bunch of stuff
    collection_summary_log(
        logger,
        goniometer,
        detector,
        attenuator,
        beam,
        source,
        timestamps,
    )

    # Get to the actual writing
    try:
        NXmx_Writer = NXmxFileWriter(
            master_file,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
            tot_num_imgs,
        )
        NXmx_Writer.write(start_time=timestamps[0])
        if pump_status is True:
            logger.info("Write pump information to file.")
            NXmx_Writer.add_NXnote(
                notes=pump_info,
                loc="/entry/source/notes",
            )
        NXmx_Writer.update_timestamps(timestamps[1], "end_time")
        NXmx_Writer.write_vds(
            vds_shape=(tot_num_imgs, *detector.detector_params.image_size),
            vds_dtype=vds_dtype,
        )
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        raise



"""
Define and store basic beamline utilities.
"""
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import List, Optional, Tuple

from dataclasses_json import DataClassJsonMixin

from nexgen.nxs_utils import Attenuator, Axis, Beam, Detector, Goniometer, Source
from nexgen.utils import Point3D


@dataclass
class PumpProbe(DataClassJsonMixin):
    """
    Define pump probe parameters.

    Args:
        pump_status (bool): Pump on/off
        pump_exposure (float, optional): Pump exposure time, in s.
        pump_delay (float, optional): Pump delay, in s.
        pump_repeat (int, optional): Repeat mode.
    """

    pump_status: bool = False
    pump_exposure: Optional[float] = None
    pump_delay: Optional[float] = None
    pump_repeat: Optional[int] = 0

    def __post_init__(self):
        if self.pump_exposure:
            self.pump_status = True


@dataclass
class BeamlineAxes:
    """Beamline specific axes for goniometer, detector and detector module."""

    gonio: List[Axis]
    det_axes: List[Axis]
    fast_axis: Point3D | Tuple[float, float, float]
    slow_axis: Point3D | Tuple[float, float, float]

    def __post_init__(self):
        if not isinstance(self.fast_axis, Point3D):
            self.fast_axis = Point3D(*self.fast_axis)
        if not isinstance(self.slow_axis, Point3D):
            self.slow_axis = Point3D(*self.slow_axis)


def collection_summary_log(
    logger: logging.Logger,
    goniometer: Goniometer,
    detector: Detector,
    attenuator: Attenuator,
    beam: Beam,
    source: Source,
    timestamps: Tuple[str],
):
    """General function to log a collection summary."""
    logger.info("--- COLLECTION SUMMARY ---")
    logger.info(source.__repr__())

    logger.info(f"Incident beam wavelength: {beam.wavelength}")
    logger.info(f"Attenuation: {attenuator.transmission}")

    logger.info(goniometer.__repr__())
    logger.info(detector.__repr__())

    logger.info(f"Recorded beam center is: {detector.beam_center}.")
    logger.info(f"Recorded exposure time: {detector.exp_time} s.")

    logger.info(f"Timestamps recorded: {timestamps}")



from __future__ import annotations

from .Axes import Axis, TransformationType
from .Detector import (
    Detector,
    DetectorModule,
    DetectorType,
    EigerDetector,
    JungfrauDetector,
    SinglaDetector,
    TristanDetector,
)
from .Goniometer import Goniometer
from .Sample import Sample
from .Source import Attenuator, Beam, Facility, Source

__all__ = [
    "Axis",
    "Goniometer",
    "Detector",
    "DetectorModule",
    "DetectorType",
    "SinglaDetector",
    "TristanDetector",
    "EigerDetector",
    "JungfrauDetector",
    "SinglaDetector",
    "Source",
    "Beam",
    "Attenuator",
    "Sample",
    "TransformationType",
    "Facility",
]



from nexgen.beamlines.SSX_chip import (
    Chip,
    compute_goniometer,
    fullchip_blocks_conversion,
    fullchip_conversion_table,
    read_chip_map,
)

test_chip = Chip(
    "testchip",
    num_steps=[20, 20],
    step_size=[0.125, 0.125],
    num_blocks=[2, 2],
    block_size=[3.175, 3.175],
    start_pos=[0.0, 0.0, 0.0],
)

test_goniometer = {"axes": ["omega", "sam_y", "sam_x", "phi"]}


def test_chip_tot_blocks():
    assert test_chip.tot_blocks() == 4


def test_chip_windows():
    assert test_chip.tot_windows_per_block() == 400
    assert test_chip.window_size() == (2.5, 2.5)


def test_chip_size():
    size = test_chip.chip_size()
    assert isinstance(size, tuple)
    assert size == (6.35, 6.35)


def test_chip_types():
    assert isinstance(test_chip.num_steps[0], int)
    assert isinstance(test_chip.step_size[0], float)
    assert isinstance(test_chip.num_blocks[0], int)
    assert isinstance(test_chip.block_size[0], float)


def test_no_chip_map_passed_returns_fullchip():
    res = read_chip_map(None, 1, 1)
    assert isinstance(res, dict)
    assert list(res.values())[0] == "fullchip"


def test_fullchip_conversion_table():
    table = fullchip_conversion_table(test_chip)
    assert len(table) == 4
    assert list(table.keys()) == ["01", "02", "03", "04"]
    assert list(table.values()) == [(0, 0), (0, 1), (1, 1), (1, 0)]


def test_fullchip_blocks_conversion():
    test_pos = {
        (0, 0): 4 * [0.0],
        (0, 1): [0.0, 3.175, 0.0, 0.0],
        (1, 1): [0.0, 5.55, 3.175, 0.0],
        (1, 0): [0.0, 2.375, 3.175, 0.0],
    }
    new_test_pos = fullchip_blocks_conversion(test_pos, test_chip)
    assert list(test_pos.values()) == list(new_test_pos.values())
    assert list(new_test_pos.keys()) == ["01", "02", "03", "04"]


def test_read_chip_map(dummy_chipmap_file):
    blocks = read_chip_map(
        dummy_chipmap_file.name, test_chip.num_blocks[0], test_chip.num_blocks[1]
    )
    assert isinstance(blocks, dict) and len(blocks) == 2
    assert list(blocks.keys()) == ["01", "04"]


def test_compute_goniometer_for_full_chip():
    starts = compute_goniometer(test_chip, full=True)
    assert len(starts) == test_chip.tot_blocks()
    assert list(starts.keys()) == [(0, 0), (0, 1), (1, 1), (1, 0)]


def test_compute_goniometer_from_chipmap_for_up_and_down_blocks(dummy_chipmap_file):
    blocks = read_chip_map(
        dummy_chipmap_file.name, test_chip.num_blocks[0], test_chip.num_blocks[1]
    )
    starts = compute_goniometer(test_chip, blocks=blocks)
    assert list(starts.keys()) == ["01", "04"]
    assert starts["01"]["sam_y"] == 0.0
    assert starts["01"]["sam_x"] == 0.0
    assert starts["01"]["direction"] == 1
    assert starts["04"]["sam_y"] == 2.375
    assert starts["04"]["sam_x"] == 3.175
    assert starts["04"]["direction"] == -1



"""
General tools useful to create NeXus format files.
"""
from __future__ import annotations

__author__ = "Diamond Light Source - Scientific Software"
__email__ = "data_analysis@diamond.ac.uk"
__version__ = "0.8.4"
__version_tuple__ = tuple(int(x) for x in __version__.split("."))



"""
Command line tool to generate NeXus files.
"""

import argparse
import glob
import logging
import sys
import time
from datetime import datetime
from pathlib import Path

import freephil
import h5py
import numpy as np

from .. import log
from ..tools.DataWriter import generate_event_files, generate_image_files
from ..tools.MetaReader import overwrite_beam, overwrite_detector
from ..utils import (
    get_filename_template,
    get_iso_timestamp,
    get_nexus_filename,
    units_of_time,
)
from . import (
    config_parser,
    demo_parser,
    detectormode_parser,
    nexus_parser,
    version_parser,
)
from .cli_utils import add_tristan_spec, call_writers, phil2dict

# Define a logger object
logger = logging.getLogger("nexgen.NeXusGenerator")

# Phil scopes
master_phil = freephil.parse(
    """
    output {
      master_filename = None
        .type = path
        .help = "Filename for master file"
    }
    input {
      datafile = None
        .type = path
        .help = "HDF5 file. For now, assumes pattern filename_%0{6}d.h5"
      coordinate_frame = *mcstas imgcif
        .type = choice
        .help = "Which coordinate system is being used to provide input vectors."
      vds_writer = None *dataset file
        .type = choice
        .help = "If not None, write vds along with external link to data in NeXus file, or create _vds.h5 file."
      snaked = False
        .type = bool
        .help = "Grid scan parameter. If True, the writer will draw a snaked grid."
    }

    include scope nexgen.command_line.nxs_phil.goniometer_scope

    include scope nexgen.command_line.nxs_phil.instrument_scope

    include scope nexgen.command_line.nxs_phil.detector_scope

    include scope nexgen.command_line.nxs_phil.module_scope

    include scope nexgen.command_line.nxs_phil.timestamp_scope
    """,
    process_includes=True,
)

demo_phil = freephil.parse(
    """
    output {
      master_filename = nexus_master.h5
        .type = path
        .help = "Filename for master file"
    }

    input {
      coordinate_frame = *mcstas imgcif
        .type = choice
        .help = "Which coordinate system is being used to provide input vectors"
      vds_writer = None *dataset file
        .type = choice
        .help = "If not None, either write a vds in the nexus file or create also a _vds.h5 file."
      snaked = False
        .type = bool
        .help = "Grid scan parameter. If True, the writer will draw a snaked grid."
    }

    include scope nexgen.command_line.nxs_phil.goniometer_scope

    include scope nexgen.command_line.nxs_phil.instrument_scope

    include scope nexgen.command_line.nxs_phil.detector_scope

    include scope nexgen.command_line.nxs_phil.module_scope
    """,
    process_includes=True,
)

meta_phil = freephil.parse(
    """
    output {
      master_filename = None
        .type = path
        .help = "Filename for master file"
    }
    input {
      metafile = None
        .type = path
        .help = "Path to _meta.h5 file for collection."
      datafile = None
        .type = path
        .help = "HDF5 file. For now, assumes pattern filename_%0{6}d.h5"
      coordinate_frame = *mcstas imgcif
        .type = choice
        .help = "Which coordinate system is being used to provide input vectors."
      vds_writer = None *dataset file
        .type = choice
        .help = "If not None, write vds along with external link to data in NeXus file, or create _vds.h5 file."
      snaked = False
        .type = bool
        .help = "Grid scan parameter. If True, the writer will draw a snaked grid."
    }

    include scope nexgen.command_line.nxs_phil.goniometer_scope

    include scope nexgen.command_line.nxs_phil.instrument_scope

    include scope nexgen.command_line.nxs_phil.detector_scope

    include scope nexgen.command_line.nxs_phil.module_scope

    include scope nexgen.command_line.nxs_phil.timestamp_scope
    """,
    process_includes=True,
)

# Parse command line arguments
parser = argparse.ArgumentParser(
    description="Generate a new NeXus file for data collection.",
    parents=[version_parser],
)
parser.add_argument("--debug", action="store_const", const=True)


# CLIs
def write_NXmx_cli(args):
    cl = master_phil.command_line_argument_interpreter()
    working_phil = master_phil.fetch(cl.process_and_fetch(args.phil_args))
    params = working_phil.extract()

    if args.show_config:
        working_phil.show(attributes_level=args.attributes_level)
        sys.exit()

    # Path to data file
    datafiles = [
        Path(d).expanduser().resolve() for d in glob.glob(params.input.datafile)
    ]

    # Get NeXus file name
    if params.output.master_filename:
        master_file = Path(params.output.master_filename).expanduser().resolve()
    else:
        master_file = get_nexus_filename(datafiles[0])

    # Start logger
    logfile = master_file.parent / "generate_nexus.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info("NeXus file writer for existing dataset.")

    # Add some information to logger
    logger.info("Create a NeXus file for %s" % datafiles[0])
    logger.info(
        "Number of experiment data files in directory, linked to the Nexus file: %d"
        % len(datafiles)
    )
    logger.info("NeXus file will be saved as %s" % master_file)

    # Load technical info from phil parser
    cf = params.input.coordinate_frame
    goniometer = params.goniometer
    detector = params.detector
    module = params.detector_module
    source = params.source
    beam = params.beam
    attenuator = params.attenuator
    timestamps = (
        get_iso_timestamp(params.start_time),
        get_iso_timestamp(params.end_time),
    )

    # If dealing with a tristan detector, add its specifications to detector scope.
    if "TRISTAN" in detector.description.upper():
        add_tristan_spec(detector, params.tristanSpec)
        logger.info("Tristan specs added to params for writing.")

    # Define images vs events based on detector mode
    logger.info(f"Data type: {detector.mode}.")
    if detector.mode == "events":
        data_type = ("events", None)
    else:
        if len(datafiles) == 1:
            with h5py.File(datafiles[0], "r") as f:
                num_images = f["data"].shape[0]
        else:
            from ..nxs_write.write_utils import find_number_of_images

            num_images = find_number_of_images(datafiles)
        data_type = ("images", num_images)
        logger.info(f"Total number of images: {num_images}")

    # Log information
    logger.info("Source information")
    logger.info(f"Facility: {source.name} - {source.type}.")
    logger.info(f"Beamline / Instrument: {source.beamline_name}")

    if timestamps[0] is not None:
        logger.info(f"Collection start time: {timestamps[0]}")
    else:
        logger.warning("No collection start time recorded.")
    if timestamps[1] is not None:
        logger.info(f"Collection end time: {timestamps[1]}")
    else:
        logger.warning("No collection end time recorded.")

    logger.info("Coordinate system: %s" % cf)
    if cf == "imgcif":
        logger.warning(
            "Input coordinate frame is imgcif. They will be converted to mcstas."
        )

    logger.info("Goniometer information")
    axes = goniometer.axes
    axis_vectors = goniometer.vectors
    for tu in zip(goniometer.types, goniometer.units):
        assert tu in (
            ("translation", "mm"),
            ("rotation", "deg"),
        ), "Appropriate axis units should be: mm for translations, deg for rotations"

    assert len(axis_vectors) == 3 * len(
        axes
    ), "Number of vectors does not match number of goniometer axes."

    for j in reversed(range(len(axes))):
        vector = axis_vectors[3 * j : 3 * j + 3]
        logger.info(
            f"Goniometer axis: {axes[j]} => {vector} ({goniometer.types[j]}) on {goniometer.depends[j]}. {goniometer.starts[j]} {goniometer.ends[j]} {goniometer.increments[j]}"
        )

    logger.info("\n")

    logger.info(
        f"Detector information: {detector.description}, {detector.detector_type}"
    )
    logger.info(
        f"Sensor made of {detector.sensor_material} x {detector.sensor_thickness}"
    )
    logger.info(f"Trusted pixels > {detector.underload} and < {detector.overload}")
    logger.info(
        f"Image is a {detector.image_size[::-1]} array of {detector.pixel_size} pixels"
    )

    logger.info("Detector axes:")
    axes = detector.axes
    axis_vectors = detector.vectors
    for tu in zip(detector.types, detector.units):
        assert tu in (
            ("translation", "mm"),
            ("rotation", "deg"),
        ), "Appropriate axis units should be: mm for translations, deg for rotations"

    assert len(axis_vectors) == 3 * len(
        axes
    ), "Number of vectors does not match number of detector axes."

    for j in range(len(axes)):
        vector = axis_vectors[3 * j : 3 * j + 3]
        logger.info(
            f"Detector axis: {axes[j]} => {vector} ({detector.types[j]}) on {detector.depends[j]}. {detector.starts[j]}"
        )

    if detector.flatfield is None:
        logger.info("No flatfield applied")
    else:
        logger.info(f"Flatfield correction data lives here {detector.flatfield}")

    if detector.pixel_mask is None:
        logger.info("No bad pixel mask for this detector")
    else:
        logger.info(f"Bad pixel mask lives here {detector.pixel_mask}")

    logger.info("Module information")
    logger.info(f"Number of modules: {module.num_modules}")
    logger.info(f"Fast axis at datum position: {module.fast_axis}")
    logger.info(f"Slow_axis at datum position: {module.slow_axis}")
    if module.module_offset == "0":
        logger.warning("module_offset field will not be written.")
    logger.info("")

    if params.pump_probe.pump_status is True:
        logger.info("Pump probe status is True, write relative metadata as NXnote.")
        pump_info = {
            "pump_exposure_time": params.pump_probe.pump_exp,
            "pump_delay": params.pump_probe.pump_delay,
        }
    else:
        pump_info = None

    logger.info("Start writing NeXus file ...")
    try:
        call_writers(
            master_file,
            datafiles,
            cf,
            data_type,
            phil2dict(goniometer.__dict__),
            phil2dict(detector.__dict__),
            phil2dict(module.__dict__),
            phil2dict(source.__dict__),
            phil2dict(beam.__dict__),
            phil2dict(attenuator.__dict__),
            timestamps=timestamps,
            notes=pump_info,
        )
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        logger.exception(err)

    logger.info("EOF\n")


def write_demo_cli(args):
    cl = demo_phil.command_line_argument_interpreter()
    working_phil = demo_phil.fetch(cl.process_and_fetch(args.phil_args))
    params = working_phil.extract()

    if args.show_config:
        working_phil.show(attributes_level=args.attributes_level)
        sys.exit()

    # Path to file
    master_file = Path(params.output.master_filename).expanduser().resolve()
    # Just in case ...
    if master_file.suffix == ".h5" and "master" not in master_file.stem:
        master_file = Path(master_file.as_posix().replace(".h5", "_master.h5"))

    # Start logger
    logfile = master_file.parent / "generate_demo.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info("Demo NeXus file writer with blank data HDF5 files.")

    # Get data file name template
    data_file_template = get_filename_template(master_file)

    # Add some information to logger
    logger.info("NeXus file will be saved as %s" % params.output.master_filename)
    logger.info("Data file(s) template: %s" % data_file_template)

    # Next: go through technical info (goniometer, detector, beamline etc ...)
    cf = params.input.coordinate_frame
    goniometer = params.goniometer
    detector = params.detector
    module = params.detector_module
    source = params.source
    beam = params.beam
    attenuator = params.attenuator

    # If dealing with a tristan detector, add its specifications to detector scope.
    if "TRISTAN" in detector.description.upper():
        add_tristan_spec(detector, params.tristanSpec)

    # Images or events ?
    if args.events is True:
        num_events = args.force if args.force else 1
        data_type = ("events", num_events)
    else:
        num_images = args.force if args.force else None
        data_type = ("images", num_images)

    # Log information
    logger.info("Data type: %s" % data_type[0])

    logger.info("Source information")
    logger.info(f"Facility: {source.name} - {source.type}.")
    logger.info(f"Beamline / Instrument: {source.beamline_name}")

    logger.info("Coordinate system: %s" % cf)
    if cf == "imgcif":
        logger.warning(
            "Input coordinate frame is imgcif. They will be converted to mcstas."
        )

    logger.info("Goniometer information")
    axes = goniometer.axes
    axis_vectors = goniometer.vectors

    for tu in zip(goniometer.types, goniometer.units):
        assert tu in (("translation", "mm"), ("rotation", "deg"))

    assert len(axis_vectors) == 3 * len(
        axes
    ), "Number of vectors does not match number of axes."

    for j in reversed(range(len(axes))):
        vector = axis_vectors[3 * j : 3 * j + 3]
        logger.info(
            f"Goniometer axis: {axes[j]} => {vector} ({goniometer.types[j]}) on {goniometer.depends[j]}. {goniometer.starts[j]} {goniometer.ends[j]} {goniometer.increments[j]}"
        )

    logger.info("")

    # Fix the number of images if not passed from command line.
    if data_type[0] == "images" and data_type[1] is None:
        raise ValueError("Missing number of images. Please specify how many to write.")

    logger.info("Detector information: %s" % detector.description)
    logger.info(
        f"Sensor made of {detector.sensor_material} x {detector.sensor_thickness}"
    )
    if data_type[0] == "images":
        logger.info(f"Trusted pixels > {detector.underload} and < {detector.overload}")
    logger.info(
        f"Image is a {detector.image_size[::-1]} array of {detector.pixel_size} pixels"
    )

    logger.info("Detector axes:")
    axes = detector.axes
    axis_vectors = detector.vectors
    for tu in zip(detector.types, detector.units):
        assert tu in (("translation", "mm"), ("rotation", "deg"))

    assert len(axis_vectors) == 3 * len(axes)

    for j in range(len(axes)):
        vector = axis_vectors[3 * j : 3 * j + 3]
        logger.info(
            f"Detector axis: {axes[j]} => {vector} ({detector.types[j]}) on {detector.depends[j]}. {detector.starts[j]}"
        )

    if detector.flatfield is None:
        logger.info("No flatfield applied")
    else:
        logger.info(f"Flatfield correction data lives here {detector.flatfield}")

    if detector.pixel_mask is None:
        logger.info("No bad pixel mask for this detector")
    else:
        logger.info(f"Bad pixel mask lives here {detector.pixel_mask}")

    logger.info("Module information")
    logger.warning(f"module_offset field setting: {module.module_offset}")
    logger.info(f"Number of modules: {module.num_modules}")
    logger.info(f"Fast axis at datum position: {module.fast_axis}")
    logger.info(f"Slow_axis at datum position: {module.slow_axis}")
    logger.info("")

    # Figure out how many files will need to be written
    logger.info("Calculating number of files to write ...")
    if data_type[0] == "events":
        # Determine the number of files. Write one file per module.
        # FIXME Either a 10M or a 2M, no other possibilities at this moment.
        n_files = 10 if "10M" in detector.description.upper() else 2
    else:
        # The maximum number of images being written each dataset is 1000
        if data_type[1] <= 1000:
            n_files = 1
        else:
            n_files = int(np.ceil(data_type[1] / 1000))

    logger.info("%d file(s) containing blank data to be written." % n_files)

    # Get datafile list
    datafiles = [
        Path(data_file_template % (n + 1)).expanduser().resolve()
        for n in range(n_files)
    ]

    logger.info("Calling data writer ...")
    # Write data files
    if data_type[0] == "images":
        generate_image_files(
            datafiles, detector.image_size, detector.description, data_type[1]
        )
    else:
        exp_time = units_of_time(detector.exposure_time)
        generate_event_files(
            datafiles, data_type[1], detector.description, exp_time.magnitude
        )

    logger.info("\n")

    # Record string with start_time
    start_time = datetime.fromtimestamp(time.time()).strftime("%A, %d. %B %Y %I:%M%p")

    if params.pump_probe.pump_status is True:
        logger.info("Pump probe status is True, write relative metadata as NXnote.")
        pump_info = {
            "pump_exposure_time": params.pump_probe.pump_exp,
            "pump_delay": params.pump_probe.pump_delay,
        }
    else:
        pump_info = None

    # Record string with end_time
    end_time = datetime.fromtimestamp(time.time()).strftime("%A, %d. %B %Y %I:%M%p")

    # Write /entry/start_time and /entry/end_time
    timestamps = (get_iso_timestamp(start_time), get_iso_timestamp(end_time))
    logger.info("Start writing demo NXmx NeXus and data files ...")
    try:
        call_writers(
            master_file,
            datafiles,
            cf,
            data_type,
            phil2dict(goniometer.__dict__),
            phil2dict(detector.__dict__),
            phil2dict(module.__dict__),
            phil2dict(source.__dict__),
            phil2dict(beam.__dict__),
            phil2dict(attenuator.__dict__),
            timestamps=timestamps,
            notes=pump_info,
        )
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        logger.exception(err)

    logger.info("EOF\n")


def write_with_meta_cli(args):
    cl = meta_phil.command_line_argument_interpreter()
    working_phil = meta_phil.fetch(cl.process_and_fetch(args.phil_args))
    params = working_phil.extract()

    if args.show_config:
        working_phil.show(attributes_level=args.attributes_level)
        sys.exit()

    # Path to meta file
    if params.input.metafile:
        metafile = Path(params.input.metafile).expanduser().resolve()
    else:
        sys.exit(
            "Please pass a _meta.h5 file. If not available use 'nexus' option instead."
        )

    # Get NeXus filename
    if params.output.master_filename:
        master_file = Path(params.output.master_filename).expanduser().resolve()
    else:
        master_file = get_nexus_filename(metafile)

    # If no datafile has been passed, look for them in the directory
    if params.input.datafile:
        datafiles = [
            Path(d).expanduser().resolve() for d in glob.glob(params.input.datafile)
        ]
    else:
        datafile_pattern = (
            metafile.parent / f"{master_file.stem}_{6*'[0-9]'}.h5"
        ).as_posix()
        datafiles = [
            Path(d).expanduser().resolve() for d in sorted(glob.glob(datafile_pattern))
        ]

    # Start logger
    logfile = master_file.parent / "generate_nexus_from_meta.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info("NeXus file writer for existing dataset with meta file available.")

    # Add some information to logger
    logger.info("Create a NeXus file for %s" % datafiles[0])
    logger.info(
        "Number of experiment data files in directory, linked to the Nexus file: %d"
        % len(datafiles)
    )
    logger.info("Meta file for the collection: %s" % metafile)
    logger.info("NeXus file will be saved as %s" % master_file)

    # Load technical info from phil parser
    cf = params.input.coordinate_frame
    goniometer = params.goniometer
    detector = params.detector
    module = params.detector_module
    source = params.source
    beam = params.beam
    attenuator = params.attenuator
    timestamps = (
        get_iso_timestamp(params.start_time),
        get_iso_timestamp(params.end_time),
    )

    # If dealing with a tristan detector, add its specifications to detector scope.
    if "TRISTAN" in detector.description.upper():
        add_tristan_spec(detector, params.tristanSpec)

    # Define images vs events based on detector mode
    logger.info(f"Data type: {detector.mode}.")
    if detector.mode == "events":
        data_type = ("events", None)
    else:
        if len(datafiles) == 1:
            with h5py.File(datafiles[0], "r") as f:
                num_images = f["data"].shape[0]
        else:
            from ..nxs_write.write_utils import find_number_of_images

            num_images = find_number_of_images(datafiles)
        data_type = ("images", num_images)
        logger.info(f"Total number of images: {num_images}")

    # Log information
    logger.info("Source information")
    logger.info(f"Facility: {source.name} - {source.type}.")
    logger.info(f"Beamline / Instrument: {source.beamline_name}")

    if timestamps[0] is not None:
        logger.info(f"Collection start time: {timestamps[0]}")
    else:
        logger.warning("No collection start time recorded.")
    if timestamps[1] is not None:
        logger.info(f"Collection end time: {timestamps[1]}")
    else:
        logger.warning("No collection end time recorded.")

    logger.info("Coordinate system: %s" % cf)
    if cf == "imgcif":
        logger.warning(
            "Input coordinate frame is imgcif. They will be converted to mcstas."
        )

    logger.info("Goniometer information")
    axes = goniometer.axes
    axis_vectors = goniometer.vectors
    for tu in zip(goniometer.types, goniometer.units):
        assert tu in (
            ("translation", "mm"),
            ("rotation", "deg"),
        ), "Appropriate axis units should be: mm for translations, deg for rotations"

    assert len(axis_vectors) == 3 * len(
        axes
    ), "Number of vectors does not match number of goniometer axes."

    for j in reversed(range(len(axes))):
        vector = axis_vectors[3 * j : 3 * j + 3]
        logger.info(
            f"Goniometer axis: {axes[j]} => {vector} ({goniometer.types[j]}) on {goniometer.depends[j]}. {goniometer.starts[j]} {goniometer.ends[j]} {goniometer.increments[j]}"
        )

    logger.info("\n")

    if detector.description is None:
        logger.error("No detector description provided, exit.")
        raise ValueError("Please provide a detector description for identification.")

    logger.info(
        f"Detector information: {detector.description}, {detector.detector_type}"
    )
    logger.info(
        f"Sensor made of {detector.sensor_material} x {detector.sensor_thickness}"
    )
    logger.info(f"Trusted pixels > {detector.underload} and < {detector.overload}")
    logger.info(
        f"Image is a {detector.image_size[::-1]} array of {detector.pixel_size} pixels"
    )

    logger.info("Detector axes:")
    axes = detector.axes
    axis_vectors = detector.vectors
    for tu in zip(detector.types, detector.units):
        assert tu in (
            ("translation", "mm"),
            ("rotation", "deg"),
        ), "Appropriate axis units should be: mm for translations, deg for rotations"

    assert len(axis_vectors) == 3 * len(
        axes
    ), "Number of vectors does not match number of detector axes."

    for j in range(len(axes)):
        vector = axis_vectors[3 * j : 3 * j + 3]
        logger.info(
            f"Detector axis: {axes[j]} => {vector} ({detector.types[j]}) on {detector.depends[j]}. {detector.starts[j]}"
        )

    if detector.flatfield is None:
        logger.info("No flatfield applied")
    else:
        logger.info(f"Flatfield correction data lives here {detector.flatfield}")

    if detector.pixel_mask is None:
        logger.info("No bad pixel mask for this detector")
    else:
        logger.info(f"Bad pixel mask lives here {detector.pixel_mask}")

    logger.info("Module information")
    logger.info(f"Number of modules: {module.num_modules}")
    logger.info(f"Fast axis at datum position: {module.fast_axis}")
    logger.info(f"Slow_axis at datum position: {module.slow_axis}")
    if module.module_offset == "0":
        logger.warning("module_offset field will not be written.")
    logger.info("")

    if args.no_ow:
        logger.warning(f"The following datasets will not be overwritten: {args.no_ow}")
        ignore = args.no_ow
    else:
        ignore = None

    logger.info("Looking through _meta.h5 file for metadata.")

    with h5py.File(metafile, "r", libver="latest", swmr=True) as mf:
        overwrite_beam(mf, detector.description, beam)
        _ = overwrite_detector(mf, detector, ignore)

    logger.info("Start writing NeXus file ...")

    if params.pump_probe.pump_status is True:
        logger.info("Pump probe status is True, write relative metadata as NXnote.")
        pump_info = {
            "pump_exposure_time": params.pump_probe.pump_exp,
            "pump_delay": params.pump_probe.pump_delay,
        }
    else:
        pump_info = None

    try:
        call_writers(
            master_file,
            datafiles,
            cf,
            data_type,
            phil2dict(goniometer.__dict__),
            phil2dict(detector.__dict__),
            phil2dict(module.__dict__),
            phil2dict(source.__dict__),
            phil2dict(beam.__dict__),
            phil2dict(attenuator.__dict__),
            metafile=True,
            timestamps=timestamps,
            notes=pump_info,
        )
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        logger.exception(err)

    logger.info("EOF\n")


# Define subparsers
subparsers = parser.add_subparsers(
    help="Choose whether to write a NXmx NeXus file for a collection or a demo. \
        Run generate_nexus <command> --help to see the parameters for each sub-command.",
    required=True,
    dest="sub-command",
)

parser_NXmx = subparsers.add_parser(
    "1",
    aliases=["nexus"],
    description=("Trigger NeXus file writing pointing to existing data."),
    parents=[nexus_parser, config_parser],
)
parser_NXmx.set_defaults(func=write_NXmx_cli)

parser_NXmx_demo = subparsers.add_parser(
    "2",
    aliases=["demo"],
    description=(
        "Trigger NeXus and blank data file writing. "
        "This option always requires either the -i or -e flags, which are mutually exclusive arguments."
    ),
    parents=[demo_parser, detectormode_parser, config_parser],
)
parser_NXmx_demo.set_defaults(func=write_demo_cli)

parser_NXmx_meta = subparsers.add_parser(
    "3",
    aliases=["meta"],
    description=(
        "Trigger NeXus file writing pointing to an existing collection with a meta file."
    ),
    parents=[nexus_parser, config_parser],
)
parser_NXmx_meta.add_argument(
    "-no",
    "--no_ow",
    nargs="+",
    help="List of datasets that should not be overwritten even if present in meta file",
    type=str,
)
parser_NXmx_meta.set_defaults(func=write_with_meta_cli)


def main():
    args = parser.parse_args()
    args.func(args)


# main()



"""
Utilities for axes definition
"""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Tuple

from ..utils import Point3D


class TransformationType(str, Enum):  # StrEnum only from Python 3.11
    """Define axis transformation type
    - ROTATION
    - TRANSLATION
    """

    ROTATION = "rotation"
    TRANSLATION = "translation"


# Define axes and scans
@dataclass
class Axis:
    """
    Define an axis object for goniometer or detector.

    Attributes:
        name (str): Axis name.
        depends (str): Name of the axis it depends on.
        transformation_type (TransformationType): Rotation or translation.
        vector (Point3D | Tuple): Axis vector.
        start_pos (float, optional): Start position of axis. Defaults to 0.0.
        increment (float, optional): Scan step size if the axis moves. Defaults to 0.0.
        num_steps (int, optional): Number of scan points. Defaults to 0.0.
        offset (Point3D | Tuple, optional): Axis offset. Defaults to (0.0, 0.0, 0.0).

    Properties:
        units (str): Defined depending on transformation type: deg or mm.
        end_pos (float): Last point recorded in a scan colletion. Calculated from start_pos, increment and num_steps, 1-indexed.
        is_scan (bool): Whether axis is a scan axis.
    """

    name: str
    depends: str
    transformation_type: TransformationType
    vector: Point3D | Tuple[float, float, float]
    start_pos: float = 0.0
    increment: float = 0.0
    num_steps: int = 0
    offset: Point3D | Tuple[float, float, float] = (0.0, 0.0, 0.0)

    def __post_init__(self):
        if type(self.vector) is Point3D:
            self.vector = (self.vector.x, self.vector.y, self.vector.z)
        if type(self.offset) is Point3D:
            self.offset = (self.offset.x, self.offset.y, self.offset.z)

    @property
    def units(self) -> str:
        if self.transformation_type == "translation":
            return "mm"
        else:
            return "deg"

    @property
    def end_pos(self) -> float:
        # Using the same system as scanspec
        return self.start_pos + self.increment * (self.num_steps - 1)

    @property
    def is_scan(self) -> bool:
        if self.increment != 0.0:
            return True
        return False



import tempfile

import h5py
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from nexgen.nxs_copy.copy_utils import (
    check_and_fix_det_axis,
    convert_scan_axis,
    h5str,
    identify_tristan_scan_axis,
    is_chipmap_in_tristan_nxs,
)
from nexgen.nxs_write.NXclassWriters import write_NXentry, write_NXnote
from nexgen.nxs_write.write_utils import create_attributes

test_chipmap = {"01": (0, 0)}


def test_h5str():
    assert h5str("string_obj") == "string_obj"
    assert h5str(b"bytes_obj") == "bytes_obj"


@pytest.fixture
def dummy_nexus_file():
    test_hdf_file = tempfile.TemporaryFile()
    test_nexus_file = h5py.File(test_hdf_file, "w")
    yield test_nexus_file


def test_identify_tristan_scan_axis(dummy_nexus_file):
    nxentry = write_NXentry(dummy_nexus_file)
    nxdata = nxentry.require_group("data")
    axis1 = nxdata.create_dataset("omega", data=(0, 0))
    create_attributes(
        axis1, ("transformation_type", "vector"), (b"rotation", [0, 0, 1])
    )
    axis2 = nxdata.create_dataset("sam_x", data=(0,))
    create_attributes(axis2, ("transformation_type",), ("translation",))
    nxdata.create_dataset("data", data=b"Some dummy data")

    ax, ax_attrs = identify_tristan_scan_axis(dummy_nexus_file)
    assert ax == "omega"
    assert len(ax_attrs) > 0
    assert ax_attrs["transformation_type"] == "rotation"
    assert_array_equal(ax_attrs["vector"], [0, 0, 1])


def test_convert_scan_axis(dummy_nexus_file):
    scan_range = np.arange(0, 2, 0.2)
    nxentry = write_NXentry(dummy_nexus_file)
    nxdata = nxentry.require_group("data")
    nxsample = nxentry.require_group("sample")
    nxtr = nxsample.require_group("transformations")
    axis = nxdata.create_dataset("omega", data=scan_range)
    create_attributes(axis, ("transformation_type", "vector"), (b"rotation", [0, 0, 1]))
    nxsample.require_group("sample_omega")
    nxsample["sample_omega"].create_dataset("omega", data=(0, 2))
    nxtr["omega"] = nxsample["sample_omega/omega"]

    convert_scan_axis(nxsample, nxdata, "omega", scan_range)
    assert_array_equal(nxsample["sample_omega/omega"], scan_range)
    assert_array_equal(nxtr["omega"], scan_range)
    assert (
        "omega_end" in nxsample["sample_omega"].keys()
        and "omega_increment_set" in nxsample["sample_omega"].keys()
    )
    assert_array_equal(nxsample["sample_omega/omega_increment_set"][()], 0.2)
    assert_array_equal(nxsample["sample_omega/omega_end"][()], scan_range + 0.2)


def test_check_and_fix_det_z(dummy_nexus_file):
    nxentry = write_NXentry(dummy_nexus_file)
    nxentry.create_dataset("instrument/detector/distance", data=b"150.")
    nxdet_z_grp = nxentry.require_group(
        "instrument/detector/transformations/detector_z"
    )
    det_z = nxdet_z_grp.create_dataset("det_z", data=b"150.")
    create_attributes(det_z, ("transformation_type", "units"), ("translation", "mm"))

    check_and_fix_det_axis(dummy_nexus_file)
    fixed_det = dummy_nexus_file[
        "/entry/instrument/detector/transformations/detector_z/det_z"
    ]
    assert_array_equal(fixed_det[()], np.array([150.0]))
    assert det_z.attrs["units"] == b"mm"
    assert det_z.attrs["transformation_type"] == b"translation"
    fixed_dist = dummy_nexus_file["/entry/instrument/detector/distance"]
    assert fixed_dist[()] == 0.150
    assert fixed_dist.attrs["units"] == b"m"


def test_find_chipmap(dummy_nexus_file):
    test_map = {"chipmap": str(test_chipmap)}
    nxentry = write_NXentry(dummy_nexus_file)
    write_NXnote(dummy_nexus_file, "/entry/source/notes", test_map)

    assert is_chipmap_in_tristan_nxs(dummy_nexus_file) is True
    assert is_chipmap_in_tristan_nxs(nxentry, loc="source/notes/chipmap") is True



"""
Logging configuration.
"""
import logging
import logging.config

import nexgen

# Logging set up
logging.getLogger("nexgen").addHandler(logging.NullHandler())

logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "class": "logging.Formatter",
            "format": "%(levelname)s - %(message)s",
        }
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "default",
            "stream": "ext://sys.stdout",
        }
    },
    "loggers": {
        "nexgen": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": True,
        }
    },
}

logging.config.dictConfig(logging_config)


def config(logfile: str = None, write_mode: str = "a", delayed: bool = False):
    """
    Configure the logging.

    Args:
        logfile (str, optional): If passed, create a file handler for the logger to write to file the log output. Defaults to None.
        write_mode (str, optional): String indicating writing mode for the output .log file. Defaults to "a".
        delayed (bool, optional): Setting for the FileHandler delay option. If true, then file opening is deferred until the first call\
            to `emit`.Defaults to False.
    """
    nexgen_logger = logging.getLogger("nexgen")
    nexgen_logger.info(f"Using nexgen {nexgen.__version__}")
    if logfile:
        fileFormatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s || %(message)s",
            datefmt="%d-%m-%Y %I:%M:%S",
        )
        FH = logging.FileHandler(
            logfile, mode=write_mode, encoding="utf-8", delay=delayed
        )
        FH.setLevel(logging.DEBUG)
        FH.setFormatter(fileFormatter)
        nexgen_logger.addHandler(FH)


class LoggingContext:
    """
    Define a basic context manager for selective logging.
    See https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging.
    """

    def __init__(self, logger, level=None):
        self.logger = logging.getLogger(logger) if isinstance(logger, str) else logger
        self.level = level

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        # implicit return of None => don't swallow exceptions
